""" A function "plot_hdf()" to plot an hdf file. (magnetogram)"""
import numpy as np
from mesh_generator.ui.psihdf import rdhdf_2d


def plot_hdf(hdf_file_path):
    p, t, br_pt = rdhdf_2d(hdf_file_path)
    # NOTE: pcolorfast and pcolormesh want the coords of pixel corners not centers --> build a "half mesh" for p & t.
    # - This means making an array that is n+1 size and has the midpoint positions of the original.
    # - I choose to clip the endpoints of the half mesh to the original bounds, vs extrapolate.
    # - see also https://matplotlib.org/api/_as_gen/matplotlib.pyplot.pcolormesh.html .
    ph = np.concatenate([[p[0]], 0.5 * (p[1:] + p[:-1]), [p[-1]]])
    th = np.concatenate([[t[0]], 0.5 * (t[1:] + t[:-1]), [t[-1]]])
    return [ph, th, br_pt]


if __name__ == "__main__":
    # Choose the files to read, point the base_path to where you put the MapPipeline folder
    hdf_file_path = "/Users/opalissan/Desktop/br_caplan_shine_fill_pt.h5"
    plot_hdf(hdf_file_path)



import numpy as np
import pyhdf.SD as h4
import h5py as h5


def rdh5(h5_filename):
    x = np.array([])
    y = np.array([])
    z = np.array([])
    f = np.array([])

    h5file = h5.File(h5_filename, 'r')
    f = h5file['Data']
    dims = f.shape
    ndims = np.ndim(f)

    # Get the scales if they exist:
    for i in range(0, ndims):
        if i == 0:
            if (len(h5file['Data'].dims[0].keys()) != 0):
                x = h5file['Data'].dims[0][0]
        elif i == 1:
            if (len(h5file['Data'].dims[1].keys()) != 0):
                y = h5file['Data'].dims[1][0]
        elif i == 2:
            if (len(h5file['Data'].dims[2].keys()) != 0):
                z = h5file['Data'].dims[2][0]

    x = np.array(x)
    y = np.array(y)
    z = np.array(z)
    f = np.array(f)

    h5file.close()

    return (x, y, z, f)


def rdhdf(hdf_filename):
    if (hdf_filename.endswith('h5')):
        x, y, z, f = rdh5(hdf_filename)
        return (x, y, z, f)

    x = np.array([])
    y = np.array([])
    z = np.array([])
    f = np.array([])

    # Open the HDF file
    sd_id = h4.SD(hdf_filename)

    # Read dataset.  In all PSI hdf4 files, the
    # data is stored in "Data-Set-2":
    sds_id = sd_id.select('Data-Set-2')
    f = sds_id.get()

    # Get number of dimensions:
    ndims = np.ndim(f)

    # Get the scales. Check if theys exist by looking at the 3rd
    # element of dim.info(). 0 = none, 5 = float32, 6 = float64.
    # see http://pysclint.sourceforge.net/pyhdf/pyhdf.SD.html#SD
    # and http://pysclint.sourceforge.net/pyhdf/pyhdf.SD.html#SDC
    for i in range(0, ndims):
        dim = sds_id.dim(i)
        if dim.info()[2] != 0:
            if i == 0:
                x = dim.getscale()
            elif i == 1:
                y = dim.getscale()
            elif i == 2:
                z = dim.getscale()

    sd_id.end()

    x = np.array(x)
    y = np.array(y)
    z = np.array(z)
    f = np.array(f)

    return (x, y, z, f)


def rdhdf_1d(hdf_filename):
    x, y, z, f = rdhdf(hdf_filename)

    return (x, f)


def rdhdf_2d(hdf_filename):
    x, y, z, f = rdhdf(hdf_filename)

    if (hdf_filename.endswith('h5')):
        return (x, y, f)
    return (y, x, f)


def rdhdf_3d(hdf_filename):
    x, y, z, f = rdhdf(hdf_filename)
    if (hdf_filename.endswith('h5')):
        return (x, y, z, f)

    return (z, y, x, f)


def wrh5(h5_filename, x, y, z, f):
    h5file = h5.File(h5_filename, 'w')

    # Create the dataset (Data is the name used by the psi data)).
    h5file.create_dataset("Data", data=f)

    # Make sure the scales are desired by checking x type, which can
    # be None or None converted by np.asarray (have to trap seperately)
    if x is None:
        x = np.array([], dtype=f.dtype)
        y = np.array([], dtype=f.dtype)
        z = np.array([], dtype=f.dtype)
    if x.any() == None:
        x = np.array([], dtype=f.dtype)
        y = np.array([], dtype=f.dtype)
        z = np.array([], dtype=f.dtype)

    # Make sure scales are the same precision as data.
    x = x.astype(f.dtype)
    y = y.astype(f.dtype)
    z = z.astype(f.dtype)

    # Get number of dimensions:
    ndims = np.ndim(f)

    # Set the scales:
    for i in range(0, ndims):
        if i == 0 and len(x) != 0:
            dim = h5file.create_dataset("dim1", data=x)
            h5file['Data'].dims.create_scale(dim, 'dim1')
            h5file['Data'].dims[0].attach_scale(dim)
            h5file['Data'].dims[0].label = 'dim1'
        if i == 1 and len(y) != 0:
            dim = h5file.create_dataset("dim2", data=y)
            h5file['Data'].dims.create_scale(dim, 'dim2')
            h5file['Data'].dims[1].attach_scale(dim)
            h5file['Data'].dims[1].label = 'dim2'
        elif i == 2 and len(z) != 0:
            dim = h5file.create_dataset("dim3", data=z)
            h5file['Data'].dims.create_scale(dim, 'dim3')
            h5file['Data'].dims[2].attach_scale(dim)
            h5file['Data'].dims[2].label = 'dim3'

    # Close the file:
    h5file.close()


def wrhdf(hdf_filename, x, y, z, f):
    if (hdf_filename.endswith('h5')):
        wrh5(hdf_filename, x, y, z, f)
        return

    # Create an HDF file
    sd_id = h4.SD(hdf_filename, h4.SDC.WRITE | h4.SDC.CREATE | h4.SDC.TRUNC)

    if f.dtype == np.float32:
        ftype = h4.SDC.FLOAT32
    elif f.dtype == np.float64:
        ftype = h4.SDC.FLOAT64

    # Create the dataset (Data-Set-2 is the name used by the psi data)).
    sds_id = sd_id.create("Data-Set-2", ftype, f.shape)

    # Get number of dimensions:
    ndims = np.ndim(f)

    # Make sure the scales are desired by checking x type, which can
    # be None or None converted by np.asarray (have to trap seperately)
    if x is None:
        x = np.array([], dtype=f.dtype)
        y = np.array([], dtype=f.dtype)
        z = np.array([], dtype=f.dtype)
    if x.any() == None:
        x = np.array([], dtype=f.dtype)
        y = np.array([], dtype=f.dtype)
        z = np.array([], dtype=f.dtype)

    # Set the scales (or don't if x is none or length zero)
    for i in range(0, ndims):
        dim = sds_id.dim(i)
        if i == 0 and len(x) != 0:
            if x.dtype == np.float32:
                stype = h4.SDC.FLOAT32
            elif x.dtype == np.float64:
                stype = h4.SDC.FLOAT64
            dim.setscale(stype, x)
        elif i == 1 and len(y) != 0:
            if y.dtype == np.float32:
                stype = h4.SDC.FLOAT32
            elif y.dtype == np.float64:
                stype = h4.SDC.FLOAT64
            dim.setscale(stype, y)
        elif i == 2 and len(z) != 0:
            if z.dtype == np.float32:
                stype = h4.SDC.FLOAT32
            elif z.dtype == np.float64:
                stype = h4.SDC.FLOAT64
            dim.setscale(stype, z)

    # Write the data:
    sds_id.set(f)

    # Close the dataset:
    sds_id.endaccess()

    # Flush and close the HDF file:
    sd_id.end()


def wrhdf_1d(hdf_filename, x, f):
    x = np.asarray(x)
    y = np.array([])
    z = np.array([])
    f = np.asarray(f)
    wrhdf(hdf_filename, x, y, z, f)


def wrhdf_2d(hdf_filename, x, y, f):
    x = np.asarray(x)
    y = np.asarray(y)
    z = np.array([])
    f = np.asarray(f)
    if (hdf_filename.endswith('h5')):
        wrhdf(hdf_filename, x, y, z, f)
        return
    wrhdf(hdf_filename, y, x, z, f)


def wrhdf_3d(hdf_filename, x, y, z, f):
    x = np.asarray(x)
    y = np.asarray(y)
    z = np.asarray(z)
    f = np.asarray(f)
    if (hdf_filename.endswith('h5')):
        wrhdf(hdf_filename, x, y, z, f)
        return
    wrhdf(hdf_filename, z, y, x, f)



"""
Automation- generate the 3D mesh from a mesh dictionary.
taking this json/dictionary info and generates the mesh WITHOUT calling a UI in the first place.
Then you can write an example that say takes automatically generated mesh requirements (e.g.
MapPipelineCD/Misc/March2012/tdm_region_info.out), builds the box dictionary and then builds the mesh.
"""
import numpy as np
from mesh_generator.bin.call_psi_mesh_tool import create_psi_mesh
from mesh_generator.src.mesh_segment import MeshSegment
from mesh_generator.src.mesh import Mesh
from mesh_generator.bin.mesh_header_template import write_mesh_header
from mesh_generator.bin.tmp_mas_template import write_tmp_mas_file
from mesh_generator.hdf.write_1d_hdf_files import write_1d_mesh_hdf
from mesh_generator.hdf.read_mesh_res import read_tmp_file
import os
import json
import shutil


class MeshRes:
    def __init__(self, save_mesh_dir, json_file_path):
        # first load the data in json file.
        self.json_file_path = json_file_path
        self.dict_data = self.read_json_file_return_dict()

        # path for parent folder.
        self.save_mesh_dir = save_mesh_dir

        self.get_mesh_theta()
        self.get_mesh_phi()
        self.get_mesh_radial()
        self.mesh_header()
        self.write_hdf_file()
        self.copy_json_file()

    def read_json_file_return_dict(self):
        """ read input json file and return dict"""
        with open(self.json_file_path, 'r') as data:
            dict_data = json.load(data)
        return dict_data

    def copy_json_file(self):
        """ copy json file to the folder. """
        shutil.copy(src=self.json_file_path, dst=os.path.join(self.save_mesh_dir, "mesh_requirements.json"))

    def mesh_header(self):
        """ write a mesh_header file """
        write_mesh_header(tmp_mas_r_file=os.path.join(self.save_mesh_dir, "tmp_mas_r.dat"),
                          tmp_mas_t_file=os.path.join(self.save_mesh_dir, "tmp_mas_t.dat"),
                          tmp_mas_p_file=os.path.join(self.save_mesh_dir, "tmp_mas_p.dat"),
                          mesh_header_file_path=os.path.join(self.save_mesh_dir, "mesh_header.dat"))

    def write_hdf_file(self):
        """ write 1d hdf file with mesh spacing."""
        r, dr, r_ratio = read_tmp_file(name_of_tmp_file=os.path.join(self.save_mesh_dir, "mesh_res_r.dat"))
        t, dt, t_ratio = read_tmp_file(name_of_tmp_file=os.path.join(self.save_mesh_dir, "mesh_res_t.dat"))
        p, dp, p_ratio = read_tmp_file(name_of_tmp_file=os.path.join(self.save_mesh_dir, "mesh_res_p.dat"))
        write_1d_mesh_hdf(outdir=self.save_mesh_dir, mesh_name="1x", r=r, t=t, p=p, fac=1.0)

    def get_mesh_theta(self):
        if 'BG_RATIO' in self.dict_data.keys():
            MeshSegment.DEFAULT_BG_REGION_RATIO = self.dict_data['BG_RATIO']
        if 'FG_RATIO' in self.dict_data.keys():
            MeshSegment.DEFAULT_FG_REGION_RATIO = self.dict_data['FG_RATIO']

        mesh = Mesh(self.dict_data['lower_bnd_t'], self.dict_data['upper_bnd_t'], False)
        for box in self.dict_data['box_list']:
            if box['t0'] is not None and box['t1'] is not None:
                if 0 <= box['t0'] < np.pi and 0 < box['t1'] <= np.pi:
                    mesh.insert_mesh_segment(MeshSegment(box['t0'], box['t1'], box['ds'],
                                                         var_ds_ratio=box['tp_var_ds']))
        if mesh.is_valid():
            adj_mesh = mesh.resolve_mesh_segments().json_dict()
            legacy_mesh = mesh.build_legacy_mesh().json_dict()
            create_psi_mesh(adj_mesh, legacy_mesh, 't', self.save_mesh_dir, output_file_name="tmp_mesh_t.dat",
                            mesh_res_file_name="mesh_res_t.dat", save_plot=True, save_plot_path=self.save_mesh_dir,
                            plot_file_name="theta_mesh_spacing.png")
            write_tmp_mas_file(mesh_type='t', tmp_mesh_dir_name=os.path.join(self.save_mesh_dir, "tmp_mesh_t.dat"),
                               tmp_mas_dir_name=os.path.join(self.save_mesh_dir, "tmp_mas_t.dat"))

    def get_mesh_phi(self):
        if 'BG_RATIO' in self.dict_data.keys():
            MeshSegment.DEFAULT_BG_REGION_RATIO = self.dict_data['BG_RATIO']
        if 'FG_RATIO' in self.dict_data.keys():
            MeshSegment.DEFAULT_FG_REGION_RATIO = self.dict_data['FG_RATIO']

        mesh = Mesh(self.dict_data["lower_bnd_p"], self.dict_data["upper_bnd_p"], True)
        for box in self.dict_data["box_list"]:
            mesh.insert_mesh_segment(MeshSegment(box["p0"], box["p1"], box["ds"], var_ds_ratio=box["tp_var_ds"]))

        if mesh.is_valid():
            adj_mesh = mesh.resolve_mesh_segments().json_dict()
            legacy_mesh = mesh.build_legacy_mesh().json_dict()
            create_psi_mesh(adj_mesh, legacy_mesh, 'p', self.save_mesh_dir, output_file_name="tmp_mesh_p.dat",
                            mesh_res_file_name="mesh_res_p.dat", save_plot=True, save_plot_path=self.save_mesh_dir,
                            plot_file_name="phi_mesh_spacing.png")
            write_tmp_mas_file(mesh_type='p', tmp_mesh_dir_name=os.path.join(self.save_mesh_dir, "tmp_mesh_p.dat"),
                               tmp_mas_dir_name=os.path.join(self.save_mesh_dir, "tmp_mas_p.dat"))

    def get_mesh_radial(self):
        if 'RADIAL_BG_RATIO' in self.dict_data.keys():
            MeshSegment.DEFAULT_BG_REGION_RATIO = self.dict_data['RADIAL_BG_RATIO']
        if 'RADIAL_FG_RATIO' in self.dict_data.keys():
            MeshSegment.DEFAULT_FG_REGION_RATIO = self.dict_data['RADIAL_FG_RATIO']

        mesh = Mesh(self.dict_data["lower_bnd_r"], self.dict_data["upper_bnd_r"], False)
        for box in self.dict_data["box_list"]:
            if box['r0'] is not None and box['r1'] is not None:
                mesh.insert_mesh_segment(MeshSegment(box['r0'], box['r1'], ds=box['radial_ds'],
                                                     var_ds_ratio=box["radial_var_ds"]))

        if mesh.is_valid():
            adj_mesh = mesh.resolve_mesh_segments().json_dict()
            legacy_mesh = mesh.build_legacy_mesh().json_dict()
            create_psi_mesh(adj_mesh, legacy_mesh, 'r', self.save_mesh_dir, output_file_name="tmp_mesh_r.dat",
                            mesh_res_file_name="mesh_res_r.dat", save_plot=True, save_plot_path=self.save_mesh_dir,
                            plot_file_name="r_mesh_spacing.png")
            write_tmp_mas_file(mesh_type='r', tmp_mesh_dir_name=os.path.join(self.save_mesh_dir, "tmp_mesh_r.dat"),
                               tmp_mas_dir_name=os.path.join(self.save_mesh_dir, "tmp_mas_r.dat"))



"""
Matplotlib tkinter example modified to plot a magnetogram with pcolorfast.

see https://matplotlib.org/gallery/user_interfaces/embedding_in_tk_sgskip.html

and https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.pcolorfast.html

NOTE:
    - This worked great (fast) on my computer with MPL 3.2.1, python 3.7.6, and tk 8.6.8 from conda.
    - BUT it worked way SLOWER on the same machine with MPL 3.2.1, python 3.8.2, and tk 8.6.10 from conda.
    - I have no idea why, but this suggests that something isn't robust and versions might be important... DOH!
"""
import tkinter

from matplotlib.backends.backend_tkagg import (
    FigureCanvasTkAgg, NavigationToolbar2Tk)
# Implement the default Matplotlib key bindings.
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure import Figure

import numpy as np

root = tkinter.Tk()
root.wm_title("Embedding in Tk")

# -----------------------------------------------------------------------------
# Begin Custom Changes
# -----------------------------------------------------------------------------
from mesh_generator.ui.psihdf import rdhdf_2d

# Choose the files to read, point the base_path to where you put the MapPipeline folder
pipeline_example_folder = '/Users/opalissan/Downloads/MagnetogramExamples'
base_path = pipeline_example_folder + '/Oct2011'

# You can really tell a speed difference loading in a giant hdf file or not
UseBigHdf = True

# This one is 2451 x 1130.
if UseBigHdf:
    hdf_file_path = base_path + '/br_example_2011_10_01_raw.hdf'
    plot_b_width = 100.

# This one is 511 x 376.
else:
    hdf_file_path = base_path + '/br_input_fb_diffused.hdf'
    plot_b_width = 20.

p, t, br_pt = rdhdf_2d(hdf_file_path)

# NOTE: pcolorfast and pcolormesh want the coords of pixel corners not centers --> build a "half mesh" for p & t.
# - This means making an array that is n+1 size and has the midpoint positions of the original.
# - I choose to clip the endpoints of the half mesh to the original bounds, vs extrapolate.
# - see also https://matplotlib.org/api/_as_gen/matplotlib.pyplot.pcolormesh.html .
ph = np.concatenate([[p[0]], 0.5*(p[1:] + p[:-1]), [p[-1]]])
th = np.concatenate([[t[0]], 0.5*(t[1:] + t[:-1]), [t[-1]]])

# Make the canvas decently large.
fig = Figure(figsize=(10, 5), dpi=100)

# Put the plot somewhere arbitrary on the canvas.
rect = [0.1, 0.1, 0.8, 0.8]
ax = fig.add_axes(rect)

# Set the x,y limits and the aspect to equal length (this will flip the theta axis in the plot as desired)
ax.set_xlim(0, np.pi*2)
ax.set_ylim(np.pi, 0.0)
ax.set_aspect('equal')

# Set the color scaling values
cmin = -plot_b_width
cmax = plot_b_width
cmap = 'RdBu'

# Compare pcolorfast vs. pcolormesh (use comments)
pcolor_object = ax.pcolorfast(ph, th, br_pt, cmap=cmap, vmin=cmin, vmax=cmax)


# Optional colorbar
fig.colorbar(pcolor_object, ax=ax)

# -----------------------------------------------------------------------------
# End Custom Changes
# -----------------------------------------------------------------------------

canvas = FigureCanvasTkAgg(fig, master=root)  # A tk.DrawingArea.
canvas.draw()

toolbar = NavigationToolbar2Tk(canvas, root)
toolbar.update()


def on_key_press(event):
    print("you pressed {}".format(event.key))
    key_press_handler(event, canvas, toolbar)


canvas.mpl_connect("key_press_event", on_key_press)

button = tkinter.Button(master=root, text="Quit", command=root.quit)

# Packing order is important. Widgets are processed sequentially and if there
# is no space left, because the window is too small, they are not displayed.
# The canvas is rather flexible in its size, so we pack it last which makes
# sure the UI controls are displayed as long as possible.
button.pack(side=tkinter.BOTTOM)
canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

tkinter.mainloop()


"""
- "run_shell_command" is the main function for calling command
  line executeables.
"""
import os
import subprocess


def run_shell_command(command: str, dirname: str, debug=False):
    """
    Execute a shell command inside python.

    Arguments:
        command: a string containing the command and its arguments.
        dirname: a string containing the path of the desired working directory.
    """
    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True, cwd=dirname)

    try:
        stdout, stderr = process.communicate(timeout=300)
        if debug:
            for line in stdout.strip().decode().splitlines():
                print(line)
            for line in stderr.strip().decode().splitlines():
                print(line)
    except subprocess.TimeoutExpired:
        process.kill()

    # check the return code here, if non-zero, print the output and raise an exception.
    if process.returncode != 0:
        print('')
        print('ERROR: run_shell_command returned a non-zero exit code!', process.returncode)
        print('  Shell command:', command)
        print('  STDOUT: ')
        for line in stdout.strip().decode().splitlines():
            print('    ' + line)
        print('  STDERR: ')
        for line in stderr.strip().decode().splitlines():
            print('    ' + line)
        raise RuntimeError('Shell command: ' + str(command) + ' returned exit code: ' + str(process.returncode))

    return process.returncode


if __name__ == "__main__":
    command = "mesh output02_mesh_t.dat -o mesh_res.txt"
    run_shell_command(command, os.getcwd(), debug=True)



""" UI Mesh python modules.
The MainWindow module: tkinter_main_app.py
The FirstWindow: upload_magnetogram.py

To start the Main UI:

from mesh_generator.ui.upload_magnetogram import MeshGeneratorUI

MeshGeneratorUI() # then call the UI first window.
"""

from mesh_generator.ui.upload_magnetogram import MeshGeneratorUI
from mesh_generator.ui.tkinter_mainapp import MeshApp
from mesh_generator.ui.auto_mesh_from_dict import MeshRes


"""Legacy Mesh Segments

The mesh legacy segments contain the segment's domain, ratio, and total number of points.
The legacy mesh segments are calculated in src/mesh_segement.py.
"""
import json
import numpy


class LegacyMeshSegment:
    def __init__(self, s0, s1, ratio, num_points):
        self.s0 = numpy.float64(s0)
        self.s1 = numpy.float64(s1)
        self.ratio = numpy.float64(ratio)
        self.num_points = num_points

    def __str__(self):
        return json.dumps(
            self.json_dict(), indent=2, default=lambda o: o.json_dict())

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def json_dict(self):
        return self.__dict__



"""
Restrict panning matplotlib plot only right and left (disable panning up and down).
"""

import matplotlib as mpl


class My_Axes(mpl.axes.Axes):
    name = "My_Axes"

    def drag_pan(self, button, key, x, y):
        mpl.axes.Axes.drag_pan(self, button, 'x', x, y)  # pretend key=='x'



"""
- "run_shell_command"
call the fortran mesh and create a txt file with the resulting mesh points.
"""
import os
from mesh_generator.bin.shell_command import run_shell_command


def call_shell_command(output_path, mesh_res_path):
    """
    - "run_shell_command"
    call the fortran mesh and create a txt file with the resulting mesh points.
    """
    # bash command
    command = "mesh " + str(output_path) + " -o " + str(mesh_res_path)
    run_shell_command(command, os.getcwd(), debug=True)



"""Write mesh results in the mesh format for MAS.
Example file phi mesh: (tmp_mas_p.dat)
  np=182
  pfrac=      0.000000E+00, 2.891111E-01, 3.295278E-01, 8.093194E-01
  dpratio=    1.704159E-02, 1.000000E+00, 9.677468E+01, 6.063566E-01
  nfpmesh=5
  phishift= 6.257346E+00

Example file theta mesh: (tmp_mas_t.dat)
  nt=141
  tfrac=      0.000000E+00, 5.043889E-01, 5.606111E-01, 1.000000E+00
  dtratio=    1.826025E-02, 1.000000E+00, 4.525929E+01
  nftmesh=5

Example file radial mesh: (tmp_mas_r.dat)
  r1=10.00
  nr=107
  rfrac=      0.000000E+00, 5.555550E-03, 3.216801E-02, 3.333333E-02, 1.000000E+00
  drratio=    1.000000E+00, 3.359895E+00, 1.000000E+00, 8.819758E+01
  nfrmesh=5
"""


def write_tmp_mas_file(mesh_type: str, tmp_mesh_dir_name: str, tmp_mas_dir_name: str):
    """Write mesh results in the mesh format for MAS."""
    with open(tmp_mesh_dir_name, 'r') as mesh_file:
        mesh_data = mesh_file.readlines()

    if mesh_type == "t":
        with open(tmp_mas_dir_name, "w") as f:
            f.write("  nt=%s\n" % int(int(mesh_data[5])+1))
            tfrac = mesh_data[9].replace("/", "")
            f.write("  tfrac=      " + tfrac)
            dtratio = mesh_data[11].replace("/", "")
            f.write("  dtratio=    " + dtratio)
            f.write("  nftmesh=" + str(mesh_data[-1]))

    if mesh_type == "p":
        with open(tmp_mas_dir_name, "w") as f:
            f.write("  np=%s\n" % int(int(mesh_data[5])+1))
            pfrac = mesh_data[9].replace("/", "")
            f.write("  pfrac=      " + pfrac)
            dpratio = mesh_data[11].replace("/", "")
            f.write("  dpratio=    " + dpratio)
            f.write("  nfpmesh=%s\n" % str(mesh_data[-1]))
            f.write("  phishift= " + "{0:.6E}".format(float(mesh_data[-3])))

    if mesh_type == "r":
        with open(tmp_mas_dir_name, "w") as f:
            f.write("  r1=%s\n" % "{:.2f}".format(float(mesh_data[7].split(",")[1][1:])))
            f.write("  nr=%s\n" % int(int(mesh_data[5])+1))
            rfrac = mesh_data[9].replace("/", "")
            f.write("  rfrac=      " + str(rfrac))
            drratio = mesh_data[11].replace("/", "")
            f.write("  drratio=    " + drratio)
            f.write("  nfrmesh=" + str(mesh_data[-1]))


if __name__ == "__main__":
    import os

    workdir = os.getcwd()
    write_tmp_mas_file("t", workdir + "/tests/py_outputs/POT3D/tmp_mesh_t.dat",
                       workdir + "/tests/py_outputs/POT3D/tmp_mas_t.dat")
    write_tmp_mas_file("p", workdir + "/tests/py_outputs/POT3D/tmp_mesh_p.dat",
                       workdir + "/tests/py_outputs/POT3D/tmp_mas_p.dat")
    write_tmp_mas_file("r", workdir + "/tests/py_outputs/POT3D/tmp_mesh_r.dat",
                       workdir + "/tests/py_outputs/POT3D/tmp_mas_r.dat")



"""calling psi_tools "mesh"

Step 4 of creating a 1D mesh:
    *create output files and call fortran "mesh" tool to create a file with mesh points.
    *check if mesh is valid and adjust the legacy mesh total number of points.
    *plot mesh results.
"""
from mesh_generator.bin.plot_mesh_res import plot_mesh_res
from mesh_generator.bin.output_dat_template import write_output_file
from mesh_generator.bin.check_mesh_valid import check_mesh_valid
from mesh_generator.bin.psi_mesh_tool import write_mesh_res_file
from mesh_generator.bin.tmp_mesh_template import rewrite_tmp_mesh_file
import os


def create_psi_mesh(adjusted_mesh: dict, legacy_mesh: dict, mesh_type: str, dir_name: str, output_file_name=None,
                    mesh_res_file_name=None, save_plot=False, show_plot=False, save_plot_path=None,
                    plot_file_name=None, input_mesh=None):
    """
    Create output files and call fortran "mesh" code to create a txt file with mesh points and respective resolution,
    and cell-to-cell ratio.
    :param input_mesh: input mesh as a dict for plotting. Optional just for plotting.
    :param plot_file_name: file name to save plot.
    :param dir_name: path to output file.
    :param mesh_res_file_name: name of file with mesh points. (t,dt, ratio) essential for plotting.
    :param output_file_name: name of file with mesh results. ex: output02_mesh_t.dat
    :param adjusted_mesh: adjusted mesh dictionary (step 2)
    :param save_plot_path: path where to save plot.
    :param save_plot: save plot as png.
    :param show_plot: bool. show matplotlib plot in interactive window.
    :param mesh_type: 't'/'p'/'r', used for plot title.
    :param legacy_mesh: legacy mesh dictionary (step 3)
    """

    # mesh output files full path
    output_path = os.path.join(dir_name, output_file_name)
    mesh_res_path = os.path.join(dir_name, mesh_res_file_name)

    # "write_output_file"
    # this function creates a file with the results of the legacy mesh so the fortran mesh will be able to read it.
    # (0 filtering)
    total_legacy_num = write_output_file(legacy_mesh, 0, mesh_type, output_path)

    # call fortran tool "mesh" and write a mesh_res file with all the mesh points, respective resolution and ratio from
    # cell to cell.
    write_mesh_res_file(mesh_res_path, output_path)

    # - "check_mesh_valid" : Check if tmp_mesh_r.dat is below user requests. Optimizes the total number of points.
    check_mesh_valid(adjusted_mesh, total_legacy_num, output_path, mesh_res_path)

    # write output file in scientific notation and single float precision.
    rewrite_tmp_mesh_file(output_path)

    if save_plot or show_plot:
        # "plot_mesh_res": this function will plot the data in mesh_res.txt
        plot_mesh_res(input_mesh=input_mesh, adjusted_mesh=adjusted_mesh, save_plot=save_plot, show_plot=show_plot,
                      mesh_res_file_name=mesh_res_path, label=mesh_type, save_plot_path=save_plot_path,
                      plot_file_name=plot_file_name)


if __name__ == "__main__":
    from tests.ar_test import *
    from mesh_generator.bin.tmp_mas_template import write_tmp_mas_file

    WorkDir = os.getcwd()
    PlotDir = "../../plots/"

    # create_psi_mesh(adjust__mesh_theta_1().json_dict(), legacy__mesh_theta_1().json_dict(), 't', WorkDir,
    #                 output_file_name="tmp_mesh_t.dat", mesh_res_file_name="mesh_res_t.dat",
    #                 save_plot=False, show_plot=True, save_plot_path=PlotDir, plot_file_name="phi_t_mesh.png")

    # write_tmp_mas_file(mesh_type="t", tmp_mesh_dir_name=WorkDir + "/tmp_mesh_t.dat",
    #                    tmp_mas_dir_name=WorkDir + "/tmp_mas_t.dat")
    #
    # print(adjust__mesh_phi_35())
    # print(legacy__mesh_phi_35())
    create_psi_mesh(adjust__mesh_phi_31().json_dict(), legacy__mesh_phi_31().json_dict(), 'p', WorkDir,
                    output_file_name="tmp_mesh_p.dat", mesh_res_file_name="mesh_res_p.dat", show_plot=True,
                    save_plot=False, input_mesh=get_mesh_phi_31().json_dict())

    # create_psi_mesh(adjust__mesh_r_1().json_dict(), legacy__mesh_r_1().json_dict(), 'r', WorkDir,
    #                     output_file_name="tmp_mesh_r.dat", mesh_res_file_name="mesh_res_r.dat",
    #                     save_plot=True, show_plot=True, save_plot_path=PlotDir, plot_file_name="radial_1_mesh.png")



"""This is a translation of the FORTRAN mesh psi_tool.
Generate a one-dimensional mesh.
Input: tmp_mesh_[t/p/r].dat (legacy mesh results)
Output: mesh_res_[t/p/r].txt (a file with mesh points).
"""
import numpy as np


def write_mesh_res_file(output_file_path, input_file_path):
    """Writes a file with mesh points, respective resolution, and cell-to-cell ratio.
    :param input_file_path: input file path.
    :param output_file_path: output file path.
    :return: mesh_res_[t/p/r].txt (a file with mesh points).
    """
    # Open the input file.
    with open(input_file_path, 'r') as f:
        input_data = f.readlines()

    # Get the coordinate label
    label = input_data[1].rstrip('\n')[0]

    # Get the flag to indicate a periodic mesh.
    periodic = input_data[3].rstrip(' \n')
    periodic = is_periodic(periodic)

    # Get the number of mesh points.
    nx = int(input_data[5].rstrip('\n'))
    # Get the domain limits.
    x0 = float(input_data[7].rstrip('\n').split(",")[0])
    x1 = float(input_data[7].rstrip('\n').split(",")[1])

    # Get the node positions.
    xfrac = input_data[9].rstrip('\n')
    xfrac = xfrac[:len(xfrac) - 2]
    xfrac = xfrac.split(",")
    xfrac = [float(item) for item in xfrac]

    # Get the magnification factors.
    dxratio = input_data[11].rstrip('\n')
    dxratio = dxratio[:len(dxratio) - 2]
    dxratio = dxratio.split(",")
    dxratio = [float(item) for item in dxratio]

    # Get the shift amount.
    xshift = float(input_data[13].rstrip('\n'))

    # Get the number of times to filter the mesh.
    nfilt = int(input_data[-1])

    # Generate the mesh.
    c = mesh_coordinate_transforms(label, nx, x0, x1, xfrac, dxratio, nfilt, periodic, xshift)

    # Write the mesh to a text file if requested.
    write_mesh(output_file_path, label, nx, periodic, c)


def mesh_coordinate_transforms(label, nc, c0, c1, frac, dratio, nfilt, periodic, c_shift):
    """Generate a 1D Mesh.
    Input arguments:
            LABEL   : [character(*)]
                      Name for the mesh coordinate (example: 'x').

            NC      : [integer]
                      Number of mesh points to load.

            C0      : [real]
                      The starting location for the coordinate.

            C1      : [real]
                      The ending location for the coordinate.
                      It is required that C1.gt.C0.


            FRAC    : [real array, dimension NSEG]
                       The normalized positions of the mesh segment
                       boundaries (as a fraction of the size of the
                       domain).  For a non-periodic mesh, the first
                       value of FRAC specified must equal 0. and the
                       last value must equal 1.  For a periodic mesh,
                       FRAC must not contain both 0. and 1., since
                       these represent the same point.

            DRATIO  : [real array, dimension NSEG]
                       The ratio of the mesh spacing at the end of a
                       segment to that at the beginning.

            NFILT   : [integer]
                      The number of times to filter the mesh-point
                      distribution array.  Set NFILT=0 if filtering
                      is not desired.  Filtering can reduce
                      discontinuities in the derivative of the mesh
                      spacing.

            PERIODIC: [logical]
                       A flag to indicate whether the mesh to be
                       generated represents a periodic coordinate.
                       If the coordinate is specified as periodic,
                       the range [C0,C1] should be the whole periodic
                       interval; the first mesh point is set at C0
                       and the last mesh point, C(NC), is set at C1.

            C_SHIFT : [real]
                       Amount by which to shift the periodic coordinate.
                       C_SHIFT is only used when PERIODIC=.true.,
                       and is ignored otherwise.  A positive C_SHIFT
                       moves the mesh points to the right.

    Output arguments:

             C       : [real array, dimension NC]
                       The locations of the mesh points.
    """

    eps = 10. ** -5  # epsilon

    # Check that the number of mesh points is valid.
    if nc < 2:
        print('### ERROR in GENMESH:')
        print('### Invalid number of mesh points specified.')
        print('### There must be at least two mesh points.')
        print('Mesh coordinate: ' + str(label) + '. Number of mesh points specified = ' + str(nc))
        return

    # Check that a positive mesh interval has been specified.
    if c0 >= c1:
        print('### ERROR in GENMESH:')
        print('### Invalid mesh interval specified.')
        print('### C1 must be greater than C0.')
        print(' Mesh coordinate: ' + str(label) + ', C0 = ' + str(c0) + ', C1 = ' + str(c1))
        return

    # Find the number of values of FRAC specified.
    nf = len(frac)

    # When no values have been specified (NF=1, the default), a uniform mesh is produced.
    if nf == 1. and frac[0] == 0.:
        cs = [c0, c1]
        r = [1]

    # Check that the specified values of FRAC are monotonically increasing.
    for i in range(1, nf):
        if frac[i] < frac[i - 1]:
            print('### ERROR in GENMESH:')
            print('### Invalid mesh specification.')
            print('Mesh coordinate: ', label)
            print('The values in FRAC must increase monotonically.')
            print('FRAC = ', frac)
            return

    # Check the specified values of FRAC.
    if periodic:
        if frac[0] < 0. or frac[-1] > 1.:
            # A periodic mesh requires the specified values of FRAC to be in the range 0. to 1.
            print('### ERROR in GENMESH:')
            print('### Invalid mesh specification.')
            print('Mesh coordinate: ', label)
            print('For a periodic coordinate, the values in FRAC must be between 0. and 1.')
            print('FRAC = ', frac)
            return

        if frac[0] == 0. and frac[-1] == 1.:
            # A periodic mesh cannot contain both 0. and 1. in FRAC, since these represent the same point.
            print('### ERROR in GENMESH:')
            print('### Invalid mesh specification.')
            print('Mesh coordinate: ', label)
            print('For a periodic coordinate, the values in FRAC must not contain both 0. and 1.')
            print('FRAC = ', frac)
            return

    if not periodic:
        # A non-periodic mesh requires the first specified value of FRAC to be 0., and the last value to equal 1.
        if frac[0] != 0.:
            print('### ERROR in GENMESH:')
            print('### Invalid mesh specification.')
            print('Mesh coordinate: ', label)
            print('For a non-periodic coordinate, the first value of FRAC must equal 0. ')
            print('FRAC = ', frac)
            return

        if frac[-1] != 1.:
            print('### ERROR in GENMESH:')
            print('### Invalid mesh specification.')
            print('Mesh coordinate: ', label)
            print('For a non-periodic coordinate, the last value of FRAC must equal 1. ')
            print('FRAC = ', frac)
            return

    # Check that the required values of DRATIO have been set, and are positive.
    if periodic:
        nr = nf
    else:
        nr = nf - 1

    for i in range(0, nr):
        if dratio[i] <= 0.:
            print('### ERROR in GENMESH:')
            print('### Invalid mesh specification.')
            print('Mesh coordinate: ', label)
            print('A required value in DRATIO has not been set or is not positive.')
            print('DRATIO = ', dratio)
            return

    # Check that an inherently discontinuous mesh has not been
    if periodic and nr == 1 and dratio[0] != 1.:
        print('### WARNING from GENMESH:')
        print('### Discontinuous mesh specification.')
        print('Mesh coordinate: ', label)
        print('An inherently discontinuous mesh has been specified.')
        print('FRAC = ', frac)
        print('DRATIO = ', dratio)

    # Set the number of segments
    ns = nf - 1

    # For a periodic coordinate, add points at XI=0. and XI=1. If they are not already present.
    if periodic:
        if frac[0] != 0.:
            ns = ns + 1
        if frac[-1] != 1.:
            ns = ns + 1

    # allocate cs and r
    cs, r = [None] * (int(ns + 1)), [None] * int(ns)

    # Set up the coordinate limits of the segments.
    if periodic:
        if frac[0] != 0.:
            # allocate cs and r
            cs[0] = c0
            cs[1:] = [c0 + (c1 - c0) * x for x in frac]
            if frac[nf - 1] != 1.:
                alpha = (1. - frac[-1]) / (frac[0] + 1. - frac[-1])
                r[0] = dratio[-1] / (1. + alpha * (dratio[-1] - 1.))
                r[1:] = dratio
                cs[-1] = c1
                r[-1] = 1. + alpha * (dratio[-1] - 1.)
            else:
                r[0] = dratio[-1]
                r[1:nr] = dratio[:nr - 1]
        else:
            cs = [c0 + (c1 - c0) * x for x in frac]
            cs.append(c1)
            r = dratio
    else:
        cs = [c0 + (c1 - c0) * x for x in frac]
        r = dratio

    # convert to numpy arrays
    cs = np.asarray(cs)
    r = np.asarray(r)

    # allocate xi, a, f
    xi = np.zeros(int(ns + 1))
    a = np.zeros(int(ns))
    f = np.zeros(int(ns))

    # Compute the XI values at the segment limits.
    for i in range(0, ns):
        dr = r[i] - 1.
        if abs(dr) < eps:
            f[i] = (cs[i + 1] - cs[i]) * (1. + 0.5 * dr)
        else:
            f[i] = (cs[i + 1] - cs[i]) * np.log(r[i]) / dr

    fac = 0.
    for i in range(ns - 1, -1, -1):
        fac = fac / r[i] + f[i]

    d = f[0] / fac
    xi[0] = 0.
    for i in range(1, ns):
        xi[i] = xi[i - 1] + d
        if i < ns - 1:
            d = d * f[i] / (f[i - 1] * r[i - 1])
    xi[ns] = 1.

    # Set the amplification factor for each segment.
    for i in range(0, ns):
        a[i] = np.log(r[i]) / (xi[i + 1] - xi[i])

    # For a periodic coordinate, find the XI shift corresponding to a shift of C_SHIFT in the coordinate.
    # Note that a positive value of C_SHIFT moves the mesh points to the right.
    if periodic:
        xi_shift = map_c_to_xi(periodic, ns, xi, cs, a, r, -c_shift, eps)
    else:
        xi_shift = 0.

    # Compute the location of the mesh points in array C by mapping from the XI values.
    dxi = 1. / (nc - 1.)
    c = np.zeros(nc)
    c[0] = c0
    for j in range(2, nc):
        xiv = (j - 1) * dxi
        c[j - 1] = map_xi_to_c(periodic, ns, xi, cs, a, r, -c_shift, xi_shift, xiv, eps)
    c[nc - 1] = c1

    # Filter the mesh if requested
    if nfilt > 0:
        for i in range(0, nfilt):
            if periodic:
                c = filter_coord_periodic(c1 - c0, nc, c)
            else:
                c = filter_coord(nc, c)
    return c


def write_mesh(fname, label, nc, periodic, c):
    """Write a one-dimensional mesh to file FNAME."""
    with open(fname, 'w') as f:
        # allocate dc and rc
        dc = c - np.roll(c, 1)
        if periodic:
            dc[0] = dc[nc - 1]
        rdc = dc / np.roll(dc, 1)
        if periodic:
            rdc[0] = rdc[nc - 1]

        f.write("i	" + label + "	d" + label + "	ratio\n")
        if periodic:
            j0 = 1
        else:
            j0 = 3
            f.write('{0: >8}'.format('1') + "\t " + "{:.14E}".format(c[0]) + "    " +
                    "{:.14E}".format(dc[1]) + "\t " + "{:.14E}".format(rdc[2]) + "\n")
            f.write('{0: >8}'.format('2') + "\t " + "{:.14E}".format(c[1]) + "\t " +
                    "{:.14E}".format(dc[1]) + "\t " + "{:.14E}".format(rdc[2]) + "\n")
        for j in range(j0, nc + 1):
            f.write('{0: >8}'.format(j) + "\t " + "{:.14E}".format(c[j - 1]) + "\t " +
                    "{:.14E}".format(dc[j - 1]) + "\t " + "{:.14E}".format(rdc[j - 1]) + "\n")


def filter_coord(n, f):
    """
    Apply a "(1,2,1)/4" low-pass digital filter to a 1D coordinate.
    The end-points F(1) and F(N) are not changed.
    """
    # Make a copy of the function.
    ff = np.copy(f)

    # Apply the filter.
    for i in range(1, n - 1):
        f[i] = 0.25 * (ff[i - 1] + 2. * ff[i] + ff[i + 1])

    return f


def filter_coord_periodic(xl, n, f):
    """
    Apply a "(1,2,1)/4" low-pass digital filter to a periodic 1D coordinate.
    XL is the periodic interval for the coordinate.
    The filtered coordinate is translated so that F(1) is preserved.
    """
    ff = np.zeros(n + 2)
    # Save the value of F(1).
    f1old = f[0]
    # Make a periodic copy of the function.
    ff[1: n + 1] = f
    ff[0] = f[int(n - 2)] - xl
    ff[n + 1] = f[1] + xl
    # Apply the filter:
    for i in range(0, n):
        f[i] = 0.25 * (ff[i] + 2. * ff[i + 1] + ff[i + 2])
    # Translate F so that F(1) is preserved.
    f1new = f[0]
    for i in range(0, n):
        f[i] = f[i] - f1new + f1old
    return f


def map_xi_to_c(periodic, ns, xi, cs, a, r, cshft, xi_shift, xiv, eps):
    """
    Get the mesh coordinate value CV for the specified xi value XIV.
    Set PERIODIC=.true. for a periodic coordinate. NS is the number of segments in the mesh definition.
    The arrays XI, CS, A, and R define the mesh mapping.
    CSHFT represents the amount to shift a periodic coordinate.
    XI_SHIFT represents the corresponding amount to shift xi.
    This is a utility routine for GENMESH.
    """
    # Find the index of the segment to which XIV belongs.
    if periodic:
        # Shift XIV by XI_SHIFT.
        xiv_p = xiv + xi_shift
        # Fold XIV_P into the main interval.
        xiv_p = fold(0., 1., xiv_p)
    else:
        xiv_p = xiv

    for i in range(0, ns):
        if xi[i] <= xiv_p <= xi[i + 1]:
            break

    d = xiv_p - xi[i]
    d1 = xi[i + 1] - xi[i]
    da = a[i] * d
    da1 = a[i] * d1

    # Interpolate the mapping function at XIV_P.
    if abs(da1) < eps:
        fac = (d * (1. + 0.5 * da)) / (d1 * (1. + 0.5 * da1))
    else:
        fac = (np.exp(da) - 1.) / (r[i] - 1.)
    cv = cs[i] + (cs[i + 1] - cs[i]) * fac

    if periodic:
        # Shift CV by the amount CSHFT.
        cv = cv - cshft
        # Fold CV into the main interval.
        cv = fold(cs[0], cs[ns], cv)
    return cv


def map_c_to_xi(periodic, ns, xi, cs, a, r, cv, eps):
    """
    Get the xi value XIV for the specified coordinate value CV.
    Set PERIODIC=.true. for a periodic coordinate.
    NS is the number of segments in the mesh definition.
    The arrays XI, CS, A, and R define the mesh mapping.
    This is a utility routine for GENMESH.
    """
    # Find the index of the segment to which CV belongs.
    if periodic:
        # Fold CV_P into the main interval.
        cv_p = fold(cs[0], cs[ns], cv)
    else:
        cv_p = cv

    for i in range(0, ns):
        if cs[i] <= cv_p <= cs[i + 1]:
            break

    d = (cv_p - cs[i]) / (cs[i + 1] - cs[i])
    da = (r[i] - 1.) * d

    # Interpolate the mapping function at XIV_P.
    if abs(da) < eps:
        fac = d * (xi[i + 1] - xi[i])
    else:
        fac = np.log(da + 1.) / a[i]

    xiv = xi[i] + fac
    return xiv


def fold(x0, x1, x):
    """
    "Fold" X into the periodic interval [X0,X1].
    On return, X is such that X0.le.X.lt.X1.
    It is assumed that X0 does not equal X1, as is physically necessary.
    If X0 and X1 are equal, the routine just returns with FOLD=X.
    :param x0: begin mesh coordinate
    :param x1: end mesh coordinate
    :param x: shift
    """
    if x0 == x1:
        return x
    xl = x1 - x0
    x = ((x - x0) % xl) + x0
    if x < x0:
        x = x + xl
    if x >= x1:
        x = x - xl
    return x


def is_periodic(state):
    if state == ".true.":
        return True
    if state == ".false.":
        return False
    else:
        return False


if __name__ == "__main__":
    import os
    import matplotlib.pyplot as plt


    def plot_mesh_res_for_vs_py(mesh_res_for_name: str, mesh_res_py_name: str):
        """
           plot the mesh points in fortran mesh vs python mesh.
           """
        xvals, yvals, rvals = [], [], []

        with open(str(mesh_res_for_name), "r") as f:
            fig, axs = plt.subplots(2, figsize=(7, 6.5))
            for idx, line in enumerate(f):
                if idx >= 1:
                    for word in line.split()[1:2]:
                        xvals.append(float(word))
                    for word in line.split()[2:3]:
                        yvals.append(float(word))
                    for word in line.split()[3:4]:
                        rvals.append(float(word))

        # plot the legacy mesh
        axs[0].scatter(xvals, yvals, label="FORTRAN mesh", s=1)
        # plot the mesh ratio
        axs[1].plot(xvals, rvals, label='FORTRAN Ratio')
        axs[1].set_title("Ratio", y=1.3)

        xvals, yvals, rvals = [], [], []

        with open(str(mesh_res_py_name), "r") as f:
            for idx, line in enumerate(f):
                if idx >= 1:
                    for word in line.split()[1:2]:
                        xvals.append(float(word))
                    for word in line.split()[2:3]:
                        yvals.append(float(word))
                    for word in line.split()[3:4]:
                        rvals.append(float(word))

        # plot the legacy mesh
        axs[0].scatter(xvals, yvals, label="Py mesh", s=1)
        # plot the mesh ratio
        axs[1].plot(xvals, rvals, label=' Py Ratio')
        axs[1].set_title("Ratio", y=1.3)

        axs[0].legend(bbox_to_anchor=(0, 1.02, 1, 0.2), loc="lower left", mode="expand", ncol=3)
        axs[1].legend(bbox_to_anchor=(0, 1.02, 1, 0.2), loc="lower left", mode="expand", ncol=1)
        plt.tight_layout()
        plt.show()


    #write_mesh_res_file(os.path.join(os.getcwd(), "mesh_res_t.dat"),
    #                    os.path.join(os.getcwd(), "tmp_mesh_t.dat"))
    plot_mesh_res_for_vs_py("mesh_res_p.txt", "../../tests/fortran_mesh_res/mesh_res_p_1.dat")



""" sub window to save the ui session with the 3d mesh requirements in a txt/json file. """
import tkinter as tk
from mesh_generator.ui.mesh_info import MeshInfo


class SaveDict:
    """
    Window for tk.tkinter_mainapp.py to save the ui_session dictionary in a txt or json file, which can later be
    uploaded again to the UI.
    """

    def __init__(self, master, ui_session, save_json_folder):
        self.ui_session = ui_session
        self.save_json_folder = save_json_folder

        self.master = master
        w = 800  # width for the Tk root
        h = 100  # height for the Tk root

        # get screen width and height
        ws = self.master.winfo_screenwidth()  # width of the screen
        hs = self.master.winfo_screenheight()  # height of the screen

        # calculate x and y coordinates for the Tk root window
        x = (ws / 2) - (w / 2)
        y = (hs / 2) - (h / 2)

        # set the dimensions of the screen
        # and where it is placed
        self.master.geometry('%dx%d+%d+%d' % (w, h, x, y))

        tk.Label(self.master, text="File path:").grid(row=0)
        self.master.title("Input json file path")
        self.e1 = tk.Entry(self.master, width=70)
        self.e1.insert(tk.END, str(self.save_json_folder) + "/mesh_requirements.json")
        self.e1.grid(row=0, column=1)

        tk.Button(self.master, text='Save', command=self.save_dict).grid(row=3, column=1, sticky=tk.W, pady=4)

    @staticmethod
    def save_session(ui_session, file_name):
        with open(file_name, 'w') as output:
            output.write(ui_session.__str__())

    def save_dict(self):
        file_name = self.e1.get()
        self.save_session(self.ui_session, file_name)
        self.master.destroy()


if __name__ == "__main__":
    import numpy as np

    my_ui_session = MeshInfo(lower_bnd_t=0, upper_bnd_t=np.pi, lower_bnd_p=0, upper_bnd_p=2 * np.pi,
                             lower_bnd_r=0, upper_bnd_r=2.5)

    root = tk.Tk()
    my_gui = SaveDict(root, my_ui_session, save_json_folder="/")
    root.mainloop()



""" Rewrite tmp_mesh_[t/p/r].dat file to be in scientific notation with single precision."""


def rewrite_tmp_mesh_file(tmp_mesh_path):
    with open(tmp_mesh_path, 'r') as file1:
        mesh_data = file1.readlines()

    # mesh domain s0, s1. Change to sn, 6 digits.
    mesh_data[7] = " " + convert_string_of_float_to_sn(mesh_data[7]) + "\n"

    # mesh segment domain.
    mesh_data[9] = mesh_data[9].replace("/", "")
    mesh_data[9] = convert_string_of_float_to_sn(mesh_data[9])
    mesh_data[9] = " " + mesh_data[9] + "/" + "\n"

    # mesh ratio.
    mesh_data[11] = mesh_data[11].replace("/", "")
    mesh_data[11] = convert_string_of_float_to_sn(mesh_data[11])
    mesh_data[11] = " " + mesh_data[11] + "/" + "\n"

    # mesh phi shift.
    mesh_data[13] = " " + "{0:.6E}".format(float(mesh_data[13])) + "\n"

    # rewrite tmp_mesh_[t/p/r].dat file.
    with open(tmp_mesh_path, 'w') as file2:
        file2.writelines(mesh_data)


def convert_string_of_float_to_sn(input_string):
    """Convert a float string to scientific notation for tmp_mas files.
    :param input_string: str.
    :return: output_string. (scientific notation with 6 digits).
    """
    str_list = input_string.split(", ")
    for ii in range(len(str_list)):
        if ii == 0:
            output_string = "{0:.6E}".format(float(str_list[ii]))
        else:
            output_string = output_string + ", " + "{0:.6E}".format(float(str_list[ii]))
    return output_string


if __name__ == "__main__":
    rewrite_tmp_mesh_file("tests/py_outputs/POT3D/tmp_mesh_r.dat")
    rewrite_tmp_mesh_file("tests/py_outputs/POT3D/tmp_mesh_t.dat")
    rewrite_tmp_mesh_file("tests/py_outputs/POT3D/tmp_mesh_p.dat")



from setuptools import setup, find_packages

with open("README_pypi.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mesh_generator",
    version="1.1.7",
    author="Predictive Science Inc",
    author_email="oissan@predsci.com",
    description="Python subroutines to create a 1D mesh in Python.",
    keywords=['Mesh Generation', 'Grid'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/predsci/mesh_generator",
    packages=find_packages(include=['mesh_generator', 'mesh_generator.src', 'mesh_generator.bin', 'mesh_generator.ui',
                                    'mesh_generator.hdf']),
    include_package_data=True,
    install_requires=['numpy>=1.15.2', 'matplotlib>=3.0.0', 'scipy >= 1.1.0', 'pyhdf>=0.9.10', 'h5py>=2.8.0'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
    python_requires='>=3.5',
)



"""Mesh

src/mesh.py: The main code to generate a MAS mesh in python.

Overview:
The process of building an optimal mesh is divided into 4 steps.
Step 1: input mesh requirements.
Step 2: adjusted_mesh: adjust mesh segments. call the function below called "resolve_mesh_segments()"
Step 3: legacy_mesh: compute the legacy mesh. Call the function below called "build_legacy_mesh()"
Step 4: call fortran mesh and write output.dat and mesh_res.txt files.

In this file, we will compute step 1, 2, and 3.
Step 1: Take in the mesh requirements specified by user and check if they are valid.
        The requirements include mesh domain, periodicity, BG_RATIO, FG_RATIO, and mesh segments.
        Functions used:
         * insert_mesh_segment() - adds mesh segments to the segment_list.
         * is_valid()

Step 2: call "resolve_mesh_segements()". This function will call the following functions (in ascending order):

        * call _resolve_mesh_segments_overlap_region():
        Check if there is overlapping in mesh segments inputted. If so it will always keep the segment with the
        lower resolution. Here is an illustration:
        mesh segments inputted     removed overlapping segments
        ||=======         ||       ||==              ||
        ||    ========    ||   ==> ||    ========    ||
        ||        ========||       ||            ====||
        ||  ==            ||       ||  ==            ||

        * call _resolve_mesh_segments_narrow_region():
        Remove all small mesh segments that are less than 4 points. And instead tag the small segment domain to the
        segment adjacent to it that has the smaller ds.

        * call _resolve_mesh_segments_ds():
        Make the current mesh continuous. Set a ds0 and ds1 for each segment.
        Meaning, ds0 is the beginning resolution and ds1 is the resolution at the end of the segment.
        Here is an illustration:
        mesh segments pre            mesh segments post
        ||==   ===        ||       ||==\\   //===\\    ||
        ||   ==           ||   ==> ||   \\==//   \\    ||
        ||            ====||       ||            \\====||
        ||                ||       ||                  ||

        * if mesh is periodic the call self.periodic_case():
        adjust the begin and last segments so the adjusted mesh will be periodic.
        Here is an illustration:
        mesh segments pre-periodic  mesh segments post-periodic
        ||\\                  ||       ||                    ||
        || \\      //==\\     ||   ==> ||         //==\\     ||
        || \\    //     \\  //||       ||\\     //     \\  //||
        ||  ====         ===  ||       ||  ====         ===  ||

Step 3: build legacy mesh. Here we take the adjusted_mesh segments and calculate the segments total number of points,
        domain (length), and ratio. The calculations to solve legacy mesh segments are in src/mesh_segments.
        Here we call build_legacy_mesh() which adds a layer of complexity since it solves the legacy mesh iteratively.
        Each time a mesh segment is *undershooting* the legacy mesh is recalculated.
"""
import json
import numpy
from .legacy_mesh import LegacyMesh
from .legacy_mesh_segment import LegacyMeshSegment
from .mesh_segment import MeshSegment

EPSILON = 1 / numpy.power(10, 5)


class Mesh:
    """
    Mesh object is the mesh in stages 1 and 2. Stage three is in legacy_mesh.py.
    The functions below include the most important functions to build a mesh.
    """

    def __init__(self, lower_bnd, upper_bnd, periodic=False, phi_shift=0.0):
        self.lower_bnd = numpy.float64(lower_bnd)
        self.upper_bnd = numpy.float64(upper_bnd)
        self.periodic = periodic
        self.phi_shift = numpy.float64(phi_shift)
        self.BG_RATIO = MeshSegment.DEFAULT_BG_REGION_RATIO
        # self.FG_RATIO = MeshSegment.DEFAULT_FG_REGION_RATIO
        self.segment_list = numpy.array([MeshSegment(self.lower_bnd, self.upper_bnd)])

    def __str__(self):
        return json.dumps(
            self.json_dict(), indent=2, default=lambda o: o.json_dict())

    def json_dict(self):
        return {
            'lower_bnd': self.lower_bnd,
            'upper_bnd': self.upper_bnd,
            'periodic': self.periodic,
            'phi_shift': self.phi_shift,
            'BG_RATIO': self.BG_RATIO,
            # 'FG_RATIO': self.FG_RATIO,
            'segment_list': [segment.json_dict() for segment in self.segment_list.tolist()]
        }

    def is_valid(self):
        """
        Check if inputted mesh is valid. Step 1 of building a mesh.
        Otherwise, print a Error message.
        """
        if self.lower_bnd >= self.upper_bnd:
            print("Error: lower_bnd >= upper_bnd.")
            return False

        if self.periodic and self.lower_bnd != 0.0:
            print("Error: If mesh is periodic, it requires lower_bnd=0.")
            return False

        if self.segment_list is None:
            print("Error: segment_list is empty. Input at least one mesh segment requirements.")
            return False

        for segment in self.segment_list:
            if segment.s0 < self.lower_bnd or segment.s1 > self.upper_bnd:
                if segment.s0 < self.lower_bnd:
                    print("Error: segment.s0 < lower_bnd.", segment.json_dict())
                if segment.s1 > self.upper_bnd:
                    print("Error: segment.s1 > upper_bnd.", segment.json_dict())
                return False
            elif not segment.segment_is_valid():
                print("Error: mesh segment is invalid.", segment.json_dict())
                return False
        return True

    def insert_mesh_segment(self, new_segment, beg_index=0):
        """
        Inserts a mesh segment into the list of mesh segments and returns its
        position in the list. The mesh segments in the list are sorted by the
        start position followed by the end position in ascending order.

        Setting `beg_index` allows the comparison and insertion to start at
        `beg_index` instead of the beginning of the list, which may speed up
        the insertion process. However, it can break the sorting order of the
        mesh segments in the list if used without care.
        """
        if new_segment.validate_bnd(s0=new_segment.s0, s1=new_segment.s1):
            if self.periodic and (new_segment.s0 < 0 or new_segment.s1 > self.upper_bnd):
                self.make_mesh_segment_periodic(new_segment)
            else:
                new_segment, segment_valid = self.force_mesh_segment_in_mesh_domain(new_segment)
                if segment_valid:
                    index = self.get_mesh_segment_index(mesh_segment=new_segment, beg_index=beg_index)
                    self.segment_list = numpy.insert(self.segment_list, index, new_segment)
        else:
            print("segment is invalid. s0 >= s1", new_segment.json_dict())

    def force_mesh_segment_in_mesh_domain(self, segment):
        """ force the non periodic mesh segments to be in the main mesh domain. """
        segment_is_valid = True
        if segment.s1 > self.upper_bnd and segment.s0 > self.upper_bnd:
            print("segment is out of main mesh bounds. ", segment.json_dict())
            segment_is_valid = False
        elif segment.s0 < self.lower_bnd and segment.s1 < self.lower_bnd:
            print("segment is out of main mesh bounds. ", segment.json_dict())
            segment_is_valid = False
        elif segment.s1 > self.upper_bnd >= segment.s0 >= self.lower_bnd:
            print("segment s1 is out of main mesh bounds. Force s1 = upper_bnd ", segment.s1)
            segment.s1 = self.upper_bnd
        elif segment.s0 < self.lower_bnd <= segment.s1 <= self.upper_bnd:
            print("segment s0 is out of main mesh bounds. Force s0 = lower_bnd ", segment.s0)
            segment.s0 = self.lower_bnd
        return segment, segment_is_valid

    def make_mesh_segment_periodic(self, new_segment):
        """ Force periodicity on mesh segments if periodic=True,
        these cases occur when: new_segment.s0 < 0 or new_segment.s1 > 2 * numpy.pi"""
        if 0 < new_segment.s0 < self.upper_bnd < new_segment.s1:
            segment_1 = MeshSegment(s0=new_segment.s0, s1=self.upper_bnd, ds=new_segment.ds0,
                                    var_ds_ratio=new_segment.var_ds_ratio)
            segment_2 = MeshSegment(s0=0, s1=new_segment.s1 - self.upper_bnd, ds=new_segment.ds0,
                                    var_ds_ratio=new_segment.var_ds_ratio)
            index1 = self.get_mesh_segment_index(mesh_segment=segment_1)
            self.segment_list = numpy.insert(self.segment_list, index1, segment_1)
            index2 = self.get_mesh_segment_index(mesh_segment=segment_2)
            self.segment_list = numpy.insert(self.segment_list, index2, segment_2)

        elif self.upper_bnd < new_segment.s0 and self.upper_bnd < new_segment.s1:
            new_segment.s0 = new_segment.s0 - self.upper_bnd
            new_segment.s1 = new_segment.s1 - self.upper_bnd
            index = self.get_mesh_segment_index(mesh_segment=new_segment)
            self.segment_list = numpy.insert(self.segment_list, index, new_segment)

        elif new_segment.s0 < 0 and new_segment.s1 < 0:
            new_segment.s0 = new_segment.s0 + self.upper_bnd
            new_segment.s1 = new_segment.s1 + self.upper_bnd
            index = self.get_mesh_segment_index(mesh_segment=new_segment)
            self.segment_list = numpy.insert(self.segment_list, index, new_segment)

        elif new_segment.s0 < 0 and 0 < new_segment.s1 < self.upper_bnd:
            segment_1 = MeshSegment(s0=0, s1=new_segment.s1, ds=new_segment.ds0, var_ds_ratio=new_segment.var_ds_ratio)
            segment_2 = MeshSegment(s0=new_segment.s0 + self.upper_bnd, s1=self.upper_bnd, ds=new_segment.ds0,
                                    var_ds_ratio=new_segment.var_ds_ratio)
            index1 = self.get_mesh_segment_index(mesh_segment=segment_1)
            self.segment_list = numpy.insert(self.segment_list, index1, segment_1)
            index2 = self.get_mesh_segment_index(mesh_segment=segment_2)
            self.segment_list = numpy.insert(self.segment_list, index2, segment_2)

        elif new_segment.s0 <= 0 and new_segment.s1 >= self.upper_bnd:
            new_segment.s0 = 0
            new_segment.s1 = self.upper_bnd
            index = self.get_mesh_segment_index(mesh_segment=new_segment)
            self.segment_list = numpy.insert(self.segment_list, index, new_segment)

    def get_mesh_segment_index(self, mesh_segment, beg_index=0):
        """ return the index of the mesh segment. """
        for index in range(beg_index, len(self.segment_list)):
            segment = self.segment_list[index]
            if mesh_segment.s0 < segment.s0:
                return index
            elif mesh_segment.s0 == segment.s0:
                if mesh_segment.s1 > segment.s1:
                    index += 1
                return index
        return len(self.segment_list)

    def remove_mesh_segment(self, index):
        """
        Removes and returns a mesh segment at the specified position in the
        list of mesh segments.
        """
        segment = self.segment_list[index]
        self.segment_list = numpy.delete(self.segment_list, index)

        return segment

    def resolve_mesh_segments(self):
        """
        Step 2 of building a mesh. Returns the adjusted mesh.
        """
        self._resolve_mesh_segments_overlap_region()  # deletes overlapping segments
        self._resolve_mesh_segments_narrow_region()  # deletes segments with 4 or less points.
        self._resolve_mesh_segments_ds()  # make adjusted mesh continuous. Every segment has a ds0 and ds1.
        if self.periodic:  # solve for periodic case by adjusting the begin and end segments.
            self._periodic_case()

        return self

    def _resolve_mesh_segments_overlap_region(self):
        """
        Resolves mesh segments with overlapping regions by modifying, removing
        and/or adding segments so that all of the mesh segments combined cover
        the entire domain without any overlapping regions.

        In the illustrations for each case and their possible results, 'a', 'b'
        and 'c' represent the regions of subsequent segments in the list. Note
        that these conditions and results remain true in the case of segment C
        starting and/or ending at the same point as segment B.
        """
        index = 0
        while index < len(self.segment_list) - 1:
            segment_a = self.segment_list[index]
            segment_b = self.segment_list[index + 1]

            if segment_a.s0 == segment_b.s0:
                if segment_a.s1 < segment_b.s1:
                    """
                    aaaa          aaaa
                    bbbbbb    =>     bbb    OR  bbbbbb
                      cccccc        cccccc        cccccc
                    """
                    if segment_a.ds0 <= segment_b.ds0:
                        self.remove_mesh_segment(index + 1)
                        segment_b.s0 = segment_a.s1
                        self.insert_mesh_segment(segment_b, index)
                    else:
                        self.remove_mesh_segment(index)
                elif segment_a.s1 == segment_b.s1:
                    """
                    aaaaaa        aaaaaa
                    bbbbbb    =>            OR  bbbbbb
                      cccccc        cccccc        cccccc
                    """
                    if segment_a.ds0 <= segment_b.ds0:
                        self.remove_mesh_segment(index + 1)
                    else:
                        self.remove_mesh_segment(index)
                else:
                    """
                    aaaaaa        aaaaaa           aaa
                    bbbb      =>            OR  bbbb
                      cccccc        cccccc        cccccc
                    """
                    if segment_a.ds0 <= segment_b.ds0:
                        self.remove_mesh_segment(index + 1)
                    else:
                        self.remove_mesh_segment(index)
                        segment_a.s0 = segment_b.s1
                        self.insert_mesh_segment(segment_a, index)
            else:
                if segment_a.s1 <= segment_b.s0:
                    """
                    aaaa           aaaa
                       bbbb    =>     bbbb
                       cccccc         cccccc
                    """
                    index += 1
                elif segment_a.s1 < segment_b.s1:
                    """
                    aaaaaa          aaaaaa          aaa
                      bbbbbb    =>       bbb    OR    bbbbbb
                        cccccc          cccccc          cccccc
                    """
                    if segment_a.ds0 <= segment_b.ds0:
                        self.remove_mesh_segment(index + 1)
                        segment_b.s0 = segment_a.s1
                        self.insert_mesh_segment(segment_b, index)
                    else:
                        segment_a.s1 = segment_b.s0
                        index += 1
                elif segment_a.s1 == segment_b.s1:
                    """
                    aaaaaa          aaaaaa          aaa
                      bbbb      =>              OR    bbbb
                        cccccc          cccccc          cccccc
                    """
                    if segment_a.ds0 <= segment_b.ds0:
                        self.remove_mesh_segment(index + 1)
                    else:
                        segment_a.s1 = segment_b.s0
                        index += 1
                else:
                    """
                    aaaaaaa          aaaaaaa          aaa aaa
                      bbb        =>               OR    bbb
                        ccccccc          ccccccc          ccccccc
                    """
                    if segment_a.ds0 <= segment_b.ds0:
                        self.remove_mesh_segment(index + 1)
                    else:
                        new_segment = MeshSegment(
                            segment_b.s1, segment_a.s1, segment_a.ds0,
                            segment_a.var_ds_ratio)
                        segment_a.s1 = segment_b.s0
                        self.insert_mesh_segment(new_segment, index + 1)
                        index += 1

    def _resolve_mesh_segments_narrow_region(self):
        """
        Removes mesh segments whose region is too narrow (4 or less points) and covers them with
        the neighboring mesh segments.
        """
        n_min = 4
        index = 0

        while index < len(self.segment_list):
            segment = self.segment_list[index]
            prev_segment = None  # default
            next_segment = None  # default
            n_test = numpy.inf  # default

            if index > 0:
                prev_segment = self.segment_list[index - 1]
            if index < len(self.segment_list) - 1:
                next_segment = self.segment_list[index + 1]

            if segment.ds0 == numpy.inf:
                if index == len(self.segment_list) - 1 and prev_segment:
                    if prev_segment.ds1 != numpy.inf:
                        n_test = numpy.floor(numpy.abs(segment.s1 - segment.s0) / prev_segment.ds1)
                elif index == 0 and next_segment:
                    if next_segment.ds0 != numpy.inf:
                        n_test = numpy.floor(numpy.abs(segment.s1 - segment.s0) / next_segment.ds0)

            elif segment.ds0 != numpy.inf:
                n_test = numpy.floor(numpy.abs(segment.s1 - segment.s0) / segment.ds0)

            # if this segment is too small:
            if n_test <= n_min:
                if index == 0:
                    next_segment.s0 = segment.s0
                elif index == (len(self.segment_list) - 1):
                    prev_segment.s1 = segment.s1
                else:
                    if prev_segment.ds0 >= next_segment.ds0:
                        if prev_segment.ds0 == numpy.inf:
                            next_segment.s0 = segment.s0
                        else:
                            prev_segment.s1 = segment.s1
                    else:
                        if next_segment.ds0 == numpy.inf:
                            prev_segment.s1 = segment.s1
                        else:
                            next_segment.s0 = segment.s0

                self.remove_mesh_segment(index)
                index += -1
            index += 1

    def _resolve_mesh_segments_ds(self):
        """
        Resolves the proper ds values at both ends of each mesh segment.

        In the illustrations for each case, '=' and '|' represent subsequent
        segments regions (prev, curr, next) and their boundaries respectively.
        The original ds values with respect to each other are shown by which
        line they are drawn (y-axis). The starting and ending ds values of a
        segment (curr) are determined by comparing its original ds value with
        its neighbors' (prev, next) original ds value respectively. Lower ds
        value always has higher priority.
        """
        # Stores the original constant ds values of each segment before they
        # are overwritten with the proper values.
        ds_list = []
        for segment in self.segment_list:
            ds_list.append(segment.ds0)

        index = 0
        end_index = len(self.segment_list) - 1
        while index <= end_index:
            segment = self.segment_list[index]

            curr_ds = ds_list[index]
            prev_ds = ds_list[index - 1] if index > 0 else curr_ds
            next_ds = ds_list[index + 1] if index < end_index else curr_ds

            if curr_ds == MeshSegment.BG_REGION_DS:
                if index == 0:
                    segment.set_ds(curr_ds, next_ds)
                elif index == end_index:
                    segment.set_ds(prev_ds, curr_ds)
                else:
                    self.resolve_gap_mid_mesh(segment, index, ds_begin=prev_ds, ds_end=next_ds)
                    if end_index == len(self.segment_list) - 2:  # if gen_resolve_gap inserted a segment
                        end_index = len(self.segment_list) - 1
                        ds_list.insert(index + 1, prev_ds)
                        ds_list[index + 1] = self.segment_list[index + 1].ds0

                        index += 1

                    else:
                        ds_list[index] = self.segment_list[index + 1].ds0

            elif curr_ds <= prev_ds:
                if curr_ds <= next_ds:
                    """
                    |==|  |==|      |==|  |  |      |  |  |==|      |  |  |  |
                    |  |==|  |  OR  |  |==|==|  OR  |==|==|  |  OR  |==|==|==|
                    |  |  |  |      |  |  |  |      |  |  |  |      |  |  |  |
                    """
                    segment.set_ds(curr_ds)
                else:
                    """
                    |==|  |  |      |  |  |  |
                    |  |==|  |  OR  |==|==|  |
                    |  |  |==|      |  |  |==|
                    """
                    segment.set_ds(curr_ds, next_ds)
            else:
                if curr_ds <= next_ds:
                    """
                    |  |  |==|      |  |  |  |
                    |  |==|  |  OR  |  |==|==|
                    |==|  |  |      |==|  |  |
                    """
                    segment.set_ds(prev_ds, curr_ds)
                else:
                    """
                    |  |  |  |
                    |  |==|  |
                    |==|  |==|
                    """
                    dsh, sh = self._check_resolve_gap_mid_mesh(index, prev_ds, next_ds)
                    if dsh - curr_ds <= EPSILON:
                        self.resolve_gap_mid_mesh(segment, index, ds_begin=prev_ds, ds_end=next_ds)
                        if end_index == len(self.segment_list) - 2:  # if gen_resolve_gap inserted a segment
                            end_index = len(self.segment_list) - 1
                            ds_list.insert(index + 1, prev_ds)
                            ds_list[index + 1] = self.segment_list[index + 1].ds0
                            index += 1
                        else:
                            ds_list[index] = self.segment_list[index + 1].ds0
                    else:
                        if segment.s0 < sh < segment.s1:
                            new_segment = MeshSegment(
                                sh, segment.s1, curr_ds, segment.var_ds_ratio)
                            new_segment.set_ds(curr_ds, next_ds)
                            segment.s1 = new_segment.s0
                            segment.set_ds(prev_ds, curr_ds)
                            index += 1
                            self.segment_list = numpy.insert(
                                self.segment_list, index, new_segment)
                            ds_list.insert(index, curr_ds)
                            end_index = len(self.segment_list) - 1

            index += 1

    def _resolve_hat_mesh_segment(self, index):
        """ This function resolves an adjusted mesh segment (step #2) where we have this pattern:
                    |  |  |  |      |  |   == |  |
                    |  |==|  | ---> |  |  //\\|  |
                    |==|  |==|      |==|//  \\|==|

        """

        segment = self.segment_list[index]

        curr_ds = segment.ds0
        prev_ds = self.segment_list[index - 1].ds1
        next_ds = self.segment_list[index + 1].ds0

        dsh, sh = self._check_resolve_gap_mid_mesh(index, prev_ds, next_ds)

        if dsh - curr_ds <= numpy.sqrt(EPSILON):
            self.resolve_gap_mid_mesh(segment, index, ds_begin=prev_ds, ds_end=next_ds)

        else:
            if segment.s0 < sh < segment.s1:
                new_segment = MeshSegment(sh, segment.s1, curr_ds, segment.var_ds_ratio)
                new_segment.set_ds(curr_ds, next_ds)
                segment.s1 = new_segment.s0
                segment.set_ds(prev_ds, curr_ds)
                index += 1
                self.segment_list = numpy.insert(self.segment_list, index, new_segment)

    def _get_ds_begin_and_ds_end(self, num_beg, num_end):
        """ return ds0 of segment begin and ds1 of segment end. This is called to check if periodic condition
        is satisfied. """
        segment_begin = self.segment_list[0]
        segment_end = self.segment_list[-1]

        ds_begin = segment_begin.ds0
        ds_end = segment_end.ds1

        if ds_begin == numpy.inf:
            # this is a fixed length segment so compute ds0 using the legacy function.
            ds_begin = segment_begin.get_ds0_using_num_points(num_beg)
        if ds_end == numpy.inf:
            # this is a fixed length segment so compute ds1 using the legacy function.
            ds_end = segment_end.get_ds1_using_num_points(num_end)

        return ds_begin, ds_end

    def build_legacy_mesh(self, max_iterations=500):
        """
        Builds and returns the legacy mesh using constraints of this mesh.
        """
        legacy_mesh = LegacyMesh(self.lower_bnd, self.upper_bnd, self.periodic, self.phi_shift)

        jj = 0  # the cumulative number of iterations.
        kk = 0  # number of iterations stuck on fixing the periodicity.
        ii = 0  # current mesh segment index.

        while ii < len(self.segment_list) and jj < max_iterations:

            segment = self.segment_list[ii]
            legacy_mesh_segment_list = segment.get_legacy_mesh_segments(self.lower_bnd, self.upper_bnd)

            if legacy_mesh_segment_list[0].s0 == segment.s0 and \
                    legacy_mesh_segment_list[-1].s1 == segment.s1:
                for legacy_segment in legacy_mesh_segment_list:
                    legacy_mesh.insert_mesh_segment(legacy_segment)

                # Update the next segment ds0 if needed OR the previous segment ds1. Ensures a continuous mesh.
                if len(legacy_mesh_segment_list) == 1:
                    if self.update_next_segment(segment, ii, legacy_segment):
                        ii = - 1
                        legacy_mesh.segment_list = numpy.array([], dtype=LegacyMeshSegment)

                if legacy_segment.s1 == self.upper_bnd and self.periodic:
                    # check if periodicity requirement is satisfied, if not adjust either begin or end segment to be
                    # fixed ds. Then, reiterate and solve for legacy mesh.
                    # Note: if kk > 2, this is a sign that we get a round off error since legacy mesh computes the
                    # number of segment points using np.ceil, this can cause the begin and end ds to be slightly off.
                    # In that case, we return the current legacy mesh and exit the while loop.
                    ds_begin, ds_end = self._get_ds_begin_and_ds_end(num_beg=legacy_mesh.segment_list[0].num_points,
                                                                     num_end=legacy_mesh.segment_list[-1].num_points)
                    if abs(ds_begin - ds_end) > EPSILON:
                        if ds_begin > ds_end:
                            self.segment_list[0].ds0 = ds_end
                        else:
                            self.segment_list[-1].ds1 = ds_begin
                        if kk < 2:
                            # this loop is caused by using np.ceil to find the legacy segment num_points.
                            ii = -1
                            legacy_mesh.segment_list = numpy.array([], dtype=LegacyMeshSegment)
                    kk += 1

            else:
                if segment == self.segment_list[-1] and \
                        segment.has_increasing_ds():
                    raise Exception("Fixed DS segment is too long for the " +
                                    "last segment. Aborting.")
                elif segment == self.segment_list[0] and \
                        segment.has_decreasing_ds():
                    raise Exception("Fixed DS segment is too long for the " +
                                    "first segment. Aborting.")

                passes_mult_segment = False
                if segment.has_increasing_ds() and segment.s1 != self.upper_bnd:

                    # Next segment
                    adjacent_segment = self.segment_list[ii + 1]

                    if legacy_mesh_segment_list[-1].s1 >= adjacent_segment.s1:
                        passes_mult_segment = True

                    if not passes_mult_segment and \
                            (not adjacent_segment.is_fixed_length() or not adjacent_segment.has_increasing_ds()):
                        # CASE: PASSES ONE SEGMENT.
                        # CASE: NOT FIXED LENGTH, NOT INCREASING DS.
                        # RECALCULATE MID HAT POINT.
                        if segment.is_fixed_ds() and adjacent_segment.is_fixed_ds() \
                                and adjacent_segment.has_decreasing_ds() and adjacent_segment.s1 != self.upper_bnd:
                            # remove one of the mesh segments.
                            self.segment_list[ii + 1].s0 = segment.s0
                            self.segment_list[ii + 1].ds1 = adjacent_segment.ds0
                            self.remove_mesh_segment(ii)
                            # recalculate the peak between the two segments.
                            self._resolve_hat_mesh_segment(index=ii + 1)

                        else:
                            segment.s1 = legacy_mesh_segment_list[-1].s1
                            adjacent_segment.s0 = legacy_mesh_segment_list[-1].s1

                    else:
                        # CASE: SEGMENT CROSSES MULTIPLE REGIONS
                        # CASE: FIXED LENGTH AND HAS INCREASING DS
                        n = segment.get_num_points_fixed_length_mesh()
                        ds = segment.get_ds1_using_num_points(n)
                        segment.ds1 = segment.BG_REGION_DS
                        adjacent_segment.ds0 = ds

                        if adjacent_segment.s1 != self.upper_bnd and adjacent_segment.is_fixed_length() \
                                and adjacent_segment.has_increasing_ds():
                            if self.segment_list[ii + 2].is_fixed_length() and \
                                    self.segment_list[ii + 2].has_decreasing_ds():
                                if self.segment_list[ii + 2].s1 != self.upper_bnd:
                                    ''''
                                        Undershoot case mid-mesh. Calculate new intersection point. 
                                        Calling resolve gap again. 
                                        ds0      ds1     ds0       ds1
                                        | |  //\\ | |    | |   //\\ | | 
                                        |=| //  \\|=| => | | //   \\|=| 
                                        | |       | |    |=|//      | |

                                    '''
                                    # print("Recalculate resolve gap.")
                                    adjacent_segment.s1 = self.segment_list[ii + 2].s1
                                    # adjacent_segment.var_ds_ratio = segment.DEFAULT_BG_REGION_RATIO
                                    adjacent_segment.ds1 = self.segment_list[ii + 2].ds1
                                    self.remove_mesh_segment(ii + 2)
                                    self.resolve_gap_mid_mesh(adjacent_segment, ii + 1, ds_begin=ds,
                                                              ds_end=adjacent_segment.ds1)

                        if adjacent_segment.s1 == self.upper_bnd and self.periodic and \
                                not adjacent_segment.has_decreasing_ds() and \
                                not self.segment_list[0].has_increasing_ds():
                            '''
                                Undershoot case end-mesh. Calculate new intersection point of resolve_gap periodic case.
                                There will now be a new phi_shift.  
                                ds0      ds1     ds0       ds1
                                | |  //\\ | |    | |   \\  | | 
                                |=| //  \\|=| => |=| // \\ | | 
                                | |       | |    |=|     \\|=|

                                '''
                            # print("Periodic case is recalculated. Last segment adjacent is undershooting.")
                            self.revert_shift()
                            self.segment_list[0].s0 = self.lower_bnd
                            self.segment_list[-1].s1 = self.upper_bnd
                            self._periodic_case()
                            legacy_mesh.phi_shift = self.phi_shift

                if segment.has_decreasing_ds() and segment.s0 != self.lower_bnd:
                    # Previous segment
                    adjacent_segment = self.segment_list[ii - 1]
                    if legacy_mesh_segment_list[-1].s0 <= adjacent_segment.s0:
                        passes_mult_segment = True

                    if not passes_mult_segment and \
                            (not adjacent_segment.is_fixed_length() or not adjacent_segment.has_decreasing_ds()):
                        # CASE: PASSES ONE SEGMENT
                        # CASE: NOT FIXED LENGTH, NOT DECREASING DS
                        # RECALCULATE MID HAT POINT.
                        if segment.is_fixed_ds() and adjacent_segment.is_fixed_ds() \
                                and adjacent_segment.has_increasing_ds() and adjacent_segment.s0 != self.lower_bnd:
                            # remove one of the mesh segments.
                            self.segment_list[ii - 1].s1 = segment.s1
                            self.segment_list[ii - 1].ds0 = adjacent_segment.ds1
                            self.remove_mesh_segment(ii)
                            # recalculate the peak between the two segments.
                            self._resolve_hat_mesh_segment(index=ii - 1)

                        else:
                            # adjust the length of the segment.
                            segment.s0 = legacy_mesh_segment_list[-1].s0
                            adjacent_segment.s1 = legacy_mesh_segment_list[-1].s0
                    else:
                        # CASE: SEGMENT CROSSES MULTIPLE REGIONS
                        # CASE: FIXED LENGTH AND HAS DECREASING DS
                        n = segment.get_num_points_fixed_length_mesh()
                        ds = segment.get_ds0_using_num_points(n)
                        segment.ds0 = segment.BG_REGION_DS
                        adjacent_segment.ds1 = ds

                        if adjacent_segment.s0 != self.lower_bnd and adjacent_segment.is_fixed_length() \
                                and adjacent_segment.has_decreasing_ds():
                            if self.segment_list[ii - 2].is_fixed_length() and \
                                    self.segment_list[ii - 2].has_increasing_ds():
                                '''
                                    Undershoot case mid-mesh. Calculate new intersection point. 
                                    Calling resolve gap again. 
                                    ds0      ds1     ds0       ds1
                                    | |  //\\ | |    | |   \\  | | 
                                    |=| //  \\|=| => |=| // \\ | | 
                                    | |       | |    |=|     \\|=|
    
                                    '''
                                # print("Recalculate resolve gap. ")
                                self.segment_list[ii - 2].s1 = adjacent_segment.s1
                                # self.segment_list[i - 2].var_ds_ratio = segment.DEFAULT_BG_REGION_RATIO
                                self.segment_list[ii - 2].ds1 = adjacent_segment.ds1
                                self.remove_mesh_segment(ii - 1)
                                self.resolve_gap_mid_mesh(self.segment_list[ii - 2], ii - 2,
                                                          ds_begin=self.segment_list[ii - 2].ds0, ds_end=ds)

                        if adjacent_segment.s0 == self.lower_bnd and self.periodic and \
                                not adjacent_segment.has_increasing_ds() and \
                                not self.segment_list[-1].has_decreasing_ds():
                            '''
                                Undershoot case begin segment. Calculate new intersection point of periodic case. 
                                There is a new phi_shift now. 
                                Calling resolve gap periodic case again. 
                                ds0      ds1     ds0       ds1
                                | |  //\\ | |    | |   \\  | | 
                                |=| //  \\|=| => |=| // \\ | | 
                                | |       | |    |=|     \\|=|

                            '''
                            # print("Periodic case is recalculated. First segment adjacent is undershooting.")
                            self.revert_shift()
                            self.segment_list[0].s0 = self.lower_bnd
                            self.segment_list[-1].s1 = self.upper_bnd
                            self._periodic_case()
                            legacy_mesh.phi_shift = self.phi_shift

                ii = -1

                legacy_mesh.segment_list = numpy.array([], dtype=LegacyMeshSegment)

            ii += 1
            jj += 1

        if jj >= max_iterations:
            raise Exception("Error: Legacy mesh failed to find a result. Try adjusting your mesh requirements. ")

        return legacy_mesh

    def update_next_segment(self, segment, i, legacy_segment):
        """
        Enforce continuity.
        If we have multiple undershoots in a row it is important to update the next segments ds0/ds1.
        Now legacy mesh will be continuous.
        """
        need_to_recalculate = False
        if segment.ds1 == numpy.inf:
            segment_ds1 = segment.ds0 * legacy_segment.ratio
            if i != len(self.segment_list) - 1:  # if there is an adjacent segment from the right.
                adjacent_segment_ds0 = self.segment_list[i + 1].ds0
                if adjacent_segment_ds0 != segment_ds1 and adjacent_segment_ds0 != numpy.inf:
                    self.segment_list[i + 1].ds0 = segment_ds1

        if segment.ds0 == numpy.inf:
            segment_ds0 = segment.ds1 / legacy_segment.ratio
            if i > 0:  # if there is an adjacent segment from the left.
                adjacent_segment_ds1 = self.segment_list[i - 1].ds1
                if round(adjacent_segment_ds1, 3) != round(segment_ds0, 3) and adjacent_segment_ds1 != numpy.inf:
                    self.segment_list[i - 1].ds1 = segment_ds0
                    need_to_recalculate = True
        return need_to_recalculate

    @staticmethod
    def get_vectors(ds_begin, ds_end, segment_begin_s1, segment_end_s0, ratio):
        """
        get the vectors to calculate the intersection of two fixed length segments.
        ds0 * R^n0 = ds1 * R^n1
        """
        ln_ratio = numpy.log(ratio)

        num_points = 1000.0
        s = numpy.arange(num_points) / (num_points - 1) * (- segment_begin_s1 + segment_end_s0) + segment_begin_s1
        n0 = numpy.log(ln_ratio * (segment_end_s0 - s) / ds_end + 1) / ln_ratio
        n1 = numpy.log(ln_ratio * (s - segment_begin_s1) / ds_begin + 1) / ln_ratio
        dsn0 = ds_end * numpy.power(ratio, n0)
        dsn1 = ds_begin * numpy.power(ratio, n1)
        return s, n0, n1, dsn0, dsn1

    @staticmethod
    def get_sh(s, dsn1, dsn0):
        """sh is the point in the s interval where the two segments intersect."""
        diff = dsn1 - dsn0
        if diff[0] > diff[len(diff) - 1]:
            diff_reverse = diff[::-1]
            s_reverse = s[::-1]
            sh = numpy.interp(0.0, diff_reverse, s_reverse)

        else:
            sh = numpy.interp(0.0, diff, s)
        return sh

    @staticmethod
    def get_n_test0(ratio0, segment_end_s0, sh, ds_end):
        """ number of points in mesh segment. Used for resolve_gap and periodic_case. """
        ln_ratio_0 = numpy.log(ratio0)
        n_test0 = numpy.ceil(numpy.log(ln_ratio_0 * (segment_end_s0 - sh) / ds_end + 1) / ln_ratio_0)
        return n_test0

    @staticmethod
    def get_n_test1(ratio1, segment_begin_s1, sh, ds_begin):
        """ number of points in mesh segment. Used for resolve_gap and periodic_case. """
        ln_ratio_1 = numpy.log(ratio1)
        n_test1 = numpy.ceil(numpy.log(ln_ratio_1 * (sh - segment_begin_s1) / ds_begin + 1) / ln_ratio_1)
        return n_test1

    def calculate_intersection(self, ds_begin, ds_end, segment_begin_s1, segment_end_s0, ratio):
        """ Find the intersection point for resolve gap. ds0 * R^n0 = ds1 * R^n1 """
        # Now find the position of the solution to the equation ds0 * R^n0 = ds1 * R^n1
        s, n0, n1, dsn0, dsn1 = self.get_vectors(ds_begin, ds_end, segment_begin_s1, segment_end_s0, ratio)

        # Find the point of the s interval where the two segments intersect.
        sh = self.get_sh(s, dsn1, dsn0)

        return s, n0, n1, dsn0, dsn1, sh

    def _check_resolve_gap_mid_mesh(self, index, ds_begin, ds_end):
        """ This is identical to the function below "resolve_gap_mid_mesh" yet it does not perform any actions.
        This is only for diagnostics. """

        segment_begin = self.segment_list[index - 1]
        segment_end = self.segment_list[index + 1]

        ratio = self.segment_list[index].var_ds_ratio
        s, n0, n1, dsn0, dsn1, sh = self.calculate_intersection(ds_begin=ds_begin, ds_end=ds_end,
                                                                segment_begin_s1=segment_begin.s1,
                                                                segment_end_s0=segment_end.s0,
                                                                ratio=ratio)

        if segment_begin.s1 < sh < segment_end.s0:
            '''
            CASE #1: the segment_begin.ds is fairly similar to segment.end_ds.
                        |  |    |  |      |  |     |  |
                        |  |    |  |  ==> |  | //\\|  |
                        |==|    |==|      |==|// \\|==|
            '''
            nh0 = numpy.ceil(numpy.interp(sh, s, n0))
            nh1 = numpy.ceil(numpy.interp(sh, s, n1))
            dsh0 = numpy.interp(sh, s, dsn0)
            dsh1 = numpy.interp(sh, s, dsn1)

            ratio0 = numpy.power((dsh0 / ds_end), (1.0 / nh0))
            ratio1 = numpy.power((dsh1 / ds_begin), (1.0 / nh1))

            # Now make sure it works
            n_test0 = self.get_n_test0(ratio0, segment_end.s0, sh, ds_end)
            n_test1 = self.get_n_test1(ratio1, segment_begin.s1, sh, ds_begin)

            if n_test0 == nh0 and n_test1 == nh1:
                ''' divide the gap segment into two segments. One segment that raises to the peak and
                another that connects the peak to segment_end.ds0'''
                return dsh0, sh
            else:
                print("Error: n_test0 and n_test1 midmesh.")

        if sh == segment_end.s0:  # end ds is much bigger than begin ds
            '''
            CASE #2: the segment_begin.ds is much smaller than segment.end_ds.
                        |  |    |==|      |  |     |==|
                        |  |    |  |  ==> |  |  // |  |
                        |==|    |  |      |==|//   |  |
            '''
            nh1 = numpy.ceil(numpy.interp(sh, s, n1))
            dsh1 = numpy.interp(sh, s, dsn1)
            ratio1 = numpy.power((dsh1 / ds_begin), (1.0 / nh1))
            # check it works
            n_test1 = self.get_n_test1(ratio1, segment_begin.s1, sh, ds_begin)
            if n_test1 == nh1:
                return dsh1, sh
            else:
                print("ERROR: n_test1 mid-mesh.")

        if sh == segment_begin.s1:
            '''
            CASE #3: the segment_begin.ds is much bigger than segment.end_ds.                  
                        |==|    |  |      |==|     |  |
                        |  |    |  |  ==> |  | \\  |  |
                        |  |    |==|      |  |   \\|==|
            '''
            nh0 = numpy.ceil(numpy.interp(sh, s, n0))
            dsh0 = numpy.interp(sh, s, dsn0)
            ratio0 = numpy.power((dsh0 / ds_end), (1.0 / nh0))
            # check it works
            n_test0 = self.get_n_test0(ratio0, segment_end.s0, sh, ds_end)
            if n_test0 == nh0:
                return dsh0, sh
            else:
                print("ERROR: n_test0 mid-mesh.")

    def resolve_gap_mid_mesh(self, segment, index, ds_begin, ds_end):
        """
        Resolves the gap between two segments "segment_begin" and "segment_end" where the segment between the two is
        called "segment_gap" which does not have a specific ds requirement.
        Hence, this function will divide "segment_gap" into two smaller segments
        such that in this gap it will go up and then down as coarsely as possible (case #1).
        If a intersection point is not possible (case #2 and case #3) it will connect begin and
        end segment such that the resulting mesh is continuous.
                    |  |    |  |      |  |     |  |
                    |  |    |  |  ==> |  | //\\|  |
                    |==|    |==|      |==|// \\|==|
        """

        segment_begin = self.segment_list[index - 1]
        segment_end = self.segment_list[index + 1]

        ratio = self.segment_list[index].var_ds_ratio
        s, n0, n1, dsn0, dsn1, sh = self.calculate_intersection(ds_begin=ds_begin, ds_end=ds_end,
                                                                segment_begin_s1=segment_begin.s1,
                                                                segment_end_s0=segment_end.s0,
                                                                ratio=ratio)

        if segment_begin.s1 < sh < segment_end.s0:
            '''
            CASE #1: the segment_begin.ds is fairly similar to segment.end_ds.
                        |  |    |  |      |  |     |  |
                        |  |    |  |  ==> |  | //\\|  |
                        |==|    |==|      |==|// \\|==|
            '''
            nh0 = numpy.ceil(numpy.interp(sh, s, n0))
            nh1 = numpy.ceil(numpy.interp(sh, s, n1))
            dsh0 = numpy.interp(sh, s, dsn0)
            dsh1 = numpy.interp(sh, s, dsn1)

            ratio0 = numpy.power((dsh0 / ds_end), (1.0 / nh0))
            ratio1 = numpy.power((dsh1 / ds_begin), (1.0 / nh1))

            # Now make sure it works
            n_test0 = self.get_n_test0(ratio0, segment_end.s0, sh, ds_end)
            n_test1 = self.get_n_test1(ratio1, segment_begin.s1, sh, ds_begin)

            if n_test0 == nh0 and n_test1 == nh1:
                ''' divide the gap segment into two segments. One segment that raises to the peak and
                another that connects the peak to segment_end.ds0'''
                self.segment_list[index].s1 = sh
                self.segment_list[index].ds0 = ds_begin
                self.segment_list[index].ds1 = segment.BG_REGION_DS
                gap_segment_2 = MeshSegment(sh, segment_end.s0)
                gap_segment_2.set_ds(segment.BG_REGION_DS, ds_end)
                self.segment_list = numpy.insert(self.segment_list, index + 1, gap_segment_2)
                self.segment_list[index].var_ds_ratio = ratio1
                self.segment_list[index + 1].var_ds_ratio = ratio0
            else:
                print("Error: n_test0 and n_test1 midmesh.")

        if sh == segment_end.s0:  # end ds is much bigger than begin ds
            '''
            CASE #2: the segment_begin.ds is much smaller than segment.end_ds.
                        |  |    |==|      |  |     |==|
                        |  |    |  |  ==> |  |  // |  |
                        |==|    |  |      |==|//   |  |
            '''
            nh1 = numpy.ceil(numpy.interp(sh, s, n1))
            dsh1 = numpy.interp(sh, s, dsn1)
            ratio1 = numpy.power((dsh1 / ds_begin), (1.0 / nh1))
            # check it works
            n_test1 = self.get_n_test1(ratio1, segment_begin.s1, sh, ds_begin)
            if n_test1 == nh1:
                self.segment_list[index].set_ds(ds_begin, segment.BG_REGION_DS)
                self.segment_list[index].var_ds_ratio = ratio1
                segment_end.set_ds(dsh1, ds_end)
            else:
                print("ERROR: n_test1 mid-mesh.")

        if sh == segment_begin.s1:
            '''
            CASE #3: the segment_begin.ds is much bigger than segment.end_ds.                  
                        |==|    |  |      |==|     |  |
                        |  |    |  |  ==> |  | \\  |  |
                        |  |    |==|      |  |   \\|==|
            '''
            nh0 = numpy.ceil(numpy.interp(sh, s, n0))
            dsh0 = numpy.interp(sh, s, dsn0)
            ratio0 = numpy.power((dsh0 / ds_end), (1.0 / nh0))
            # check it works
            n_test0 = self.get_n_test0(ratio0, segment_end.s0, sh, ds_end)
            if n_test0 == nh0:
                self.segment_list[index].set_ds(segment.BG_REGION_DS, ds_end)
                self.segment_list[index].var_ds_ratio = ratio0
                segment_begin.set_ds(ds_begin, dsh0)
            else:
                print("ERROR: n_test0 mid-mesh.")

    def solve_periodic_case_both_fixed_length(self, segment_begin, segment_end, ratio):
        """ Solve the periodic case for when both begin and end segments are fixed length. Here we use the
        same logic as resolve gap mid mesh.
        Hence, this function will tag segment begin to the end of the mesh. Now the mesh is shifted.
        "segment_gap" is the domain of segment end and segment begin, such that in this gap it will go up and
        then down as coarsely as possible (case #1).
        If a intersection point is not possible (case #2 and case #3) it will connect begin and
        end segment such that the resulting mesh is continuous.
        After finding sh and dsh we will adjust the position of segment begin and segment end which will result in a
        phi_shift. """
        s, n1, n0, dsn1, dsn0, sh = self.calculate_intersection(ds_begin=segment_end.ds0, ds_end=segment_begin.ds1,
                                                                segment_begin_s1=segment_end.s0,
                                                                segment_end_s0=2 * numpy.pi + segment_begin.s1,
                                                                ratio=ratio)

        if segment_end.s0 < sh < 2 * numpy.pi + segment_begin.s1:
            '''
            CASE #1: the segment_begin.ds1 is fairly similar to segment_end.ds0.
                        |  |    |  |      |  |     |  |
                        |  |    |  |  ==> |  | //\\|  |
                        |==|    |==|      |==|// \\|==|
            '''
            # Now discretize it by choosing the n0 and n1s and then slightly modifying
            # the ratios. This way you can guarantee that the mesh solution fits exactly
            # in this space.
            nh0 = numpy.ceil(numpy.interp(sh, s, n0))
            nh1 = numpy.ceil(numpy.interp(sh, s, n1))
            dsh = numpy.interp(sh, s, dsn1)

            ratio0 = numpy.power((dsh / segment_end.ds0), (1.0 / nh0))
            ratio1 = numpy.power((dsh / segment_begin.ds1), (1.0 / nh1))

            # Now make sure it works
            n_test0 = self.get_n_test0(ratio0, -segment_end.s0, -sh, segment_end.ds0)
            n_test1 = self.get_n_test1(ratio1, -(segment_begin.s1 + 2 * numpy.pi), -sh, segment_begin.ds1)

            # Fill in the first segment
            if n_test0 == nh0 and n_test1 == nh1:
                segment_end.s1 = sh
                segment_end.ds1 = segment_end.BG_REGION_DS  # will later be dsh
                segment_end.var_ds_ratio = ratio0
                segment_begin.s0 = (sh - 2 * numpy.pi)
                segment_begin.ds0 = segment_begin.BG_REGION_DS  # will later be dsh
                segment_begin.var_ds_ratio = ratio1
            else:
                print("Error: nh1 and nh0")

        if sh == 2 * numpy.pi + segment_begin.s1:  # begin ds is much bigger than end ds
            '''
            CASE #2: the segment_begin.ds1 is much bigger than segment_end.ds0.
                        |  |    |==|      |  |     |==|
                        |  |    |  |  ==> |  |  // |  |
                        |==|    |  |      |==|//   |  |
            '''
            nh0 = numpy.ceil(numpy.interp(sh, s, n0))
            dsh = numpy.interp(sh, s, dsn0)
            ratio0 = numpy.power((dsh / segment_end.ds0), (1.0 / nh0))
            n_test0 = self.get_n_test0(ratio0, -segment_end.s0, -sh, segment_end.ds0)
            if n_test0 == nh0:
                segment_end.ds1 = segment_end.BG_REGION_DS
                segment_end.s1 = 2 * numpy.pi + segment_begin.s1
                segment_end.var_ds_ratio = ratio0
                self.segment_list[1].ds0 = dsh
                self.remove_mesh_segment(0)
            else:
                print("Error: ntest0 != nh0 in periodic case. ")

        if sh == segment_end.s0:
            '''
            CASE #3: the segment_begin.ds1 is much smaller than segment_end.ds0.                  
                        |==|    |  |      |==|     |  |
                        |  |    |  |  ==> |  | \\  |  |
                        |  |    |==|      |  |   \\|==|
            '''
            nh1 = numpy.ceil(numpy.interp(sh, s, n1))
            dsh = numpy.interp(sh, s, dsn1)
            ratio1 = numpy.power((dsh / segment_begin.ds1), (1.0 / nh1))
            n_test1 = self.get_n_test1(ratio1, -(segment_begin.s1 + 2 * numpy.pi), -sh, segment_begin.ds1)
            if n_test1 == nh1:
                segment_begin.ds0 = segment_begin.BG_REGION_DS
                segment_begin.s0 = segment_end.s0 - 2 * numpy.pi
                segment_begin.var_ds_ratio = 1 / ratio1
                self.remove_mesh_segment(-1)
                self.segment_list[-1].ds1 = dsh
            else:
                print("Error: ntest1 != nh1 in periodic case. ")

    def _periodic_case(self):
        self._resolve_gap()
        self.remove_shift()

    def _resolve_gap(self):
        """
        solve for periodic case. 9 permutations.
        Assumes the segment begin is constant or decreasing ds and end segment is constant or increasing ds.
        Flow chart is found in mkdocs static site in Q.
        """
        segment_begin = self.segment_list[0]
        segment_end = self.segment_list[-1]

        if (segment_begin.is_constant() and segment_end.is_fixed_length()) or \
                (segment_begin.is_fixed_ds() and segment_end.is_fixed_length()):
            ''' adjust end segment to be fixed ds.'''

            segment_end.ds1 = segment_begin.ds0

        elif segment_begin.is_fixed_length() and segment_end.is_constant() or \
                segment_begin.is_fixed_length() and segment_end.is_fixed_ds():
            ''' adjust segment begin to be fixed ds.'''
            segment_begin.ds0 = segment_end.ds1

        elif (segment_begin.is_constant() and segment_end.is_constant()) or \
                (segment_begin.is_constant() and segment_end.is_fixed_ds()) or \
                (segment_begin.is_fixed_ds() and segment_end.is_constant()):
            ''' adjust to the smallest ds between the two: segment_begin.ds0 or segment_end.ds1. '''
            if segment_begin.ds0 < segment_end.ds1:
                segment_end.ds1 = segment_begin.ds0
            if segment_end.ds1 < segment_begin.ds0:
                segment_begin.ds0 = segment_end.ds1

        elif segment_begin.is_fixed_length() and segment_end.is_fixed_length() and \
                segment_begin.ds0 > segment_begin.ds1 and segment_end.ds0 < segment_end.ds1:
            '''
            Requirements: The first segment in the list should be decreasing and
            fixed length. The last segment should be increasing and fixed length.
            if sh is in the interval we add a phi_shift. 
                        ds0     ds1      ds0        ds1       
                        | |=====| |      | |========| | 
                        | |     | |      | |        | |
                        | |     | |      | |0 //\\ 1| |
                        | |     | | == > |=| //  \\ | |
                        |=|     |=|      | |//    \\|=|
            '''
            # Now find the position of the solution to the equation ds0 * R^n0 = ds1 * R^n1
            ratio = min(segment_begin.var_ds_ratio, segment_end.var_ds_ratio)
            self.solve_periodic_case_both_fixed_length(segment_begin, segment_end, ratio)

        elif segment_begin.is_fixed_ds() and segment_end.is_fixed_ds():
            ''' First and last segments are fixed ds. Then we got to make sure the fixed length 
                segment does not go above the user ds requests. If so, a constant segment will be added.'''

            if segment_begin.ds0 == segment_end.ds1:
                ''' Modify both begin and end segments. (add a phi_shift)'''
                ratio = min(segment_begin.var_ds_ratio, segment_end.var_ds_ratio)
                max_ds = max(segment_end.ds1, segment_begin.ds0)
                s, n1, n0, dsn1, dsn0, sh = self.calculate_intersection(ds_begin=segment_end.ds0,
                                                                        ds_end=segment_begin.ds1,
                                                                        segment_begin_s1=segment_end.s0,
                                                                        segment_end_s0=2 * numpy.pi + segment_begin.s1,
                                                                        ratio=ratio)
                dsh1 = numpy.interp(sh, s, dsn1)
                dsh0 = numpy.interp(sh, s, dsn0)

                if dsh1 > max_ds and dsh0 > max_ds:
                    '''
                    Add a constant segment to optimize the mesh and still remain below user request.
                                ds0     ds1      ds0        ds1
                                | |     | |      | |0 ==== 1| |
                                | |     | | == > | | //  \\ | |
                                |=|     |=|      |=|//    \\|=|
                    '''
                    s, n1, n0, dsn1, dsn0 = self.get_vectors(ds_begin=segment_end.ds0, ds_end=segment_begin.ds1,
                                                             segment_begin_s1=segment_end.s0,
                                                             segment_end_s0=2 * numpy.pi + segment_begin.s1,
                                                             ratio=ratio)

                    # Find sh0 (where dsn0 == max_ds)
                    sh0 = self.get_sh(s, max_ds, dsn0)

                    # Find sh1 (where dsn1 == max_ds)
                    sh1 = self.get_sh(s, max_ds, dsn1)

                    # Now discretize it by choosing the n0 and n1s and then slightly modifying
                    # the ratios. This way you can guarantee that the mesh solution fits exactly
                    # in this space.
                    nh0 = numpy.ceil(numpy.interp(sh0, s, n0))
                    dsh0 = numpy.interp(sh0, s, dsn0)
                    nh1 = numpy.ceil(numpy.interp(sh1, s, n1))
                    dsh1 = numpy.interp(sh1, s, dsn1)
                    ratio0 = numpy.power((dsh0 / segment_end.ds0), (1.0 / nh0))
                    ratio1 = numpy.power((dsh1 / segment_begin.ds1), (1.0 / nh1))

                    # Now make sure it works
                    n_test0 = self.get_n_test0(ratio0, -segment_end.s0, -sh0, segment_end.ds0)
                    n_test1 = self.get_n_test1(ratio1, -(segment_begin.s1 + 2 * numpy.pi), -sh1, segment_begin.ds1)

                    if n_test0 == nh0 and n_test1 == nh1:
                        segment_end.s1 = sh0
                        segment_end.ds1 = dsh0  # segment_end.BG_REGION_DS  # Approximated at dsh0
                        segment_end.var_ds_ratio = ratio0
                        gap_constant_segment = MeshSegment(sh0, sh1)
                        gap_constant_segment.set_ds(dsh0, dsh1)
                        self.segment_list = numpy.insert(self.segment_list, len(self.segment_list),
                                                         gap_constant_segment)
                        segment_begin.s0 = (sh1 - 2 * numpy.pi)
                        segment_begin.ds0 = segment_end.BG_REGION_DS  # Approximated at dsh1
                        segment_begin.var_ds_ratio = ratio1
                    else:
                        print("Error: n_test0 or n_test1")

                else:
                    '''
                    Handle like the fixed length segments, which adds a phi_shift.
                                ds0     ds1      ds0       ds1
                                | |=====| |      | |========| |
                                | |     | |      | |        | |
                                | |     | |      | |0 //\\ 1| |
                                | |     | | == > |=| //  \\ | |
                                |=|     |=|      | |//    \\|=|
                    '''
                    self.solve_periodic_case_both_fixed_length(segment_begin=segment_begin, segment_end=segment_end,
                                                               ratio=ratio)

            elif segment_begin.ds0 < segment_end.ds1:
                # ONLY ADJUST THE LAST SEGMENT SIMILAR TO THE CALCULATIONS IN RESOLVE_GAP
                s, n0, n1, dsn0, dsn1, sh = self.calculate_intersection(ds_begin=segment_end.ds0,
                                                                        ds_end=segment_begin.ds0,
                                                                        segment_begin_s1=segment_end.s0,
                                                                        segment_end_s0=segment_end.s1,
                                                                        ratio=segment_end.var_ds_ratio)
                if sh == segment_end.s0:
                    '''
                    CASE #1: the segment_end.ds0 is much smaller or bigger than segment_begin.ds0.
                    Handle like constant case. Force segment end to end as segment begin ds0. 
                    '''
                    segment_end.ds1 = segment_begin.ds0

                if segment_end.s0 < sh < segment_end.s1:
                    '''
                    CASE #2: the segment_end.ds0 is fairly similar to segment_end_ds1.
                    Segment end:
                    '''
                    nh0 = numpy.ceil(numpy.interp(sh, s, n0))
                    nh1 = numpy.ceil(numpy.interp(sh, s, n1))
                    dsh0 = numpy.interp(sh, s, dsn0)
                    dsh1 = numpy.interp(sh, s, dsn1)

                    if dsh0 < segment_end.ds1 and dsh1 < segment_end.ds1:
                        ''' Then can be treated as resolve gap. '''
                        ratio0 = numpy.power((dsh0 / segment_begin.ds0), (1.0 / nh0))
                        ratio1 = numpy.power((dsh1 / segment_end.ds0), (1.0 / nh1))

                        # Now make sure it works
                        n_test0 = self.get_n_test0(ratio0, segment_end.s1, sh, segment_begin.ds0)
                        n_test1 = self.get_n_test1(ratio1, segment_end.s0, sh, segment_end.ds0)

                        if n_test0 == nh0 and n_test1 == nh1:
                            '''
                            ds0    ds1       ds0     ds1
                            | |    | |      | |1    0| |
                            |=|    | | == > |=|// \\ | |
                            | |    |=|      | |    \\|=|
                            '''
                            gap_segment = MeshSegment(sh, segment_end.s1)
                            segment_end.s1 = sh
                            segment_end.ds1 = segment_end.BG_REGION_DS
                            gap_segment.set_ds(segment_end.BG_REGION_DS, segment_begin.ds0)
                            self.segment_list = numpy.insert(self.segment_list, len(self.segment_list), gap_segment)
                            segment_end.var_ds_ratio = ratio1
                            self.segment_list[-1].var_ds_ratio = ratio0
                        else:
                            print("Error: ntest0 != nh0 and ntest1 !=nh1")
                    else:
                        ''' 
                        Overshoot case need a constant segment added.
                                ds0     ds1      ds0       ds1
                                | |     | |      | |1     0| |        
                                | |     | |      | | ====  | |
                                |=|     | | == > |=|//  \\ | |
                                | |     |=|      | |     \\|=|
                        '''
                        # Find sh1 (where dsn1 == ds_max)
                        sh1 = self.get_sh(s, dsn1, segment_end.ds1)

                        # Find sh0 (where dsn0 == ds_max)
                        sh0 = self.get_sh(s, dsn0, segment_end.ds1)

                        # Now discretize it by choosing the n0 and n1s and then slightly modifying
                        # the ratios. This way you can guarantee that the mesh solution fits exactly
                        # in this space.
                        nh0 = numpy.ceil(numpy.interp(sh0, s, n0))
                        dsh0 = numpy.interp(sh0, s, dsn0)
                        nh1 = numpy.ceil(numpy.interp(sh1, s, n1))
                        dsh1 = numpy.interp(sh1, s, dsn1)
                        ratio0 = numpy.power((dsh0 / segment_begin.ds0), (1.0 / nh0))
                        ratio1 = numpy.power((dsh1 / segment_end.ds0), (1.0 / nh1))

                        # Now make sure it works
                        n_test0 = self.get_n_test0(ratio0, segment_end.s1, sh0, segment_begin.ds0)
                        n_test1 = self.get_n_test1(ratio1, segment_end.s0, sh1, segment_end.ds0)

                        # save segment_end.s1 used later to adjust the segments added.
                        s1_end = segment_end.s1
                        if n_test0 == nh0 and n_test1 == nh1:
                            segment_end.s1 = sh1
                            segment_end.ds1 = dsh0  # segment_end.BG_REGION_DS
                            segment_end.var_ds_ratio = ratio1
                            gap_constant_segment = MeshSegment(sh1, sh0)
                            gap_constant_segment.set_ds(dsh0, dsh1)
                            self.segment_list = numpy.insert(self.segment_list, len(self.segment_list),
                                                             gap_constant_segment)
                            gap_segement_decrease = MeshSegment(sh0, s1_end)
                            gap_segement_decrease.set_ds(dsh1, segment_begin.ds0)
                            self.segment_list = numpy.insert(self.segment_list, len(self.segment_list),
                                                             gap_segement_decrease)
                            self.segment_list[-1].var_ds_ratio = ratio0
                        else:
                            print("Error: n_test0 and n_test1")

            elif segment_begin.ds0 > segment_end.ds1 and segment_begin.ds0 != segment_begin.BG_REGION_DS:
                # ONLY ADJUST THE BEGIN SEGMENT SIMILAR TO THE CALCULATIONS IN RESOLVE_GAP
                s, n0, n1, dsn0, dsn1, sh = self.calculate_intersection(ds_begin=segment_end.ds1,
                                                                        ds_end=segment_begin.ds1,
                                                                        segment_begin_s1=segment_begin.s0,
                                                                        segment_end_s0=segment_begin.s1,
                                                                        ratio=segment_begin.var_ds_ratio)
                if sh == segment_begin.s1:
                    '''
                    CASE #1: the segment_end.ds1 is much smaller/bigger than segment_begin.ds1.
                    Handle like constant segment case. Force the begin segment to begin as segment_end.ds1.
                    Ensures periodicity.  
                    '''
                    segment_begin.ds0 = segment_end.ds1

                if segment_begin.s0 < sh < segment_begin.s1:
                    '''
                    CASE #2: the segment_begin.ds1 is fairly similar to segment_end_ds1.
                    '''
                    nh0 = numpy.ceil(numpy.interp(sh, s, n0))
                    nh1 = numpy.ceil(numpy.interp(sh, s, n1))
                    dsh0 = numpy.interp(sh, s, dsn0)
                    dsh1 = numpy.interp(sh, s, dsn1)

                    if dsh0 < segment_begin.ds0 and dsh1 < segment_begin.ds0:
                        ''''
                            ds0     ds1     ds0     ds1
                            | |    | |      | |1   0| |
                            | |    |=|  ==> | | //\\|=|
                            |=|    | |      |=|//   | |
                
                        '''
                        ratio0 = numpy.power((dsh0 / segment_begin.ds1), (1.0 / nh0))
                        ratio1 = numpy.power((dsh1 / segment_end.ds1), (1.0 / nh1))

                        # Now make sure it works
                        n_test0 = self.get_n_test0(ratio0, segment_begin.s1, sh, segment_begin.ds1)
                        n_test1 = self.get_n_test1(ratio1, segment_begin.s0, sh, segment_end.ds1)

                        if n_test0 == nh0 and n_test1 == nh1:
                            gap_segment = MeshSegment(sh, segment_begin.s1)
                            gap_segment.set_ds(segment_begin.BG_REGION_DS, segment_begin.ds1)
                            segment_begin.ds1 = segment_begin.BG_REGION_DS
                            segment_begin.ds0 = segment_end.ds1
                            segment_begin.s1 = sh
                            self.segment_list = numpy.insert(self.segment_list, 1, gap_segment)
                            segment_begin.var_ds_ratio = ratio1
                            self.segment_list[1].var_ds_ratio = ratio0
                        else:
                            print("Error: n_test0 and n_test1")

                    else:
                        ''''
                            Overshoot case. Need to add a constant segment. 
                            ds0      ds1      ds0       ds1
                            | |      | |      | |1 === 0| |
                            | |      |=|  ==> | | //  \\|=|
                            |=|      | |      |=|//     | |

                        '''

                        # Find sh1 (where dsn1 == ds_max)
                        sh1 = self.get_sh(s, dsn1, segment_begin.ds0)

                        # Find sh0 (where dsn0 == ds_max)
                        sh0 = self.get_sh(s, dsn0, segment_begin.ds0)

                        # Now discretize it by choosing the n0 and n1s and then slightly modifying
                        # the ratios. This way you can guarantee that the mesh solution fits exactly
                        # in this space.
                        nh0 = numpy.ceil(numpy.interp(sh0, s, n0))
                        dsh0 = numpy.interp(sh0, s, dsn0)
                        nh1 = numpy.ceil(numpy.interp(sh1, s, n1))
                        dsh1 = numpy.interp(sh1, s, dsn1)
                        ratio0 = numpy.power((dsh0 / segment_begin.ds1), (1.0 / nh0))
                        ratio1 = numpy.power((dsh1 / segment_end.ds1), (1.0 / nh1))

                        # Now make sure it works
                        n_test0 = self.get_n_test0(ratio0, segment_begin.s1, sh0, segment_begin.ds1)
                        n_test1 = self.get_n_test1(ratio1, segment_begin.s0, sh1, segment_end.ds1)

                        s1_begin = segment_begin.s1
                        ds1_begin = segment_begin.ds1

                        if n_test1 == nh1 and n_test0 == nh0:
                            segment_begin.s1 = sh1
                            segment_begin.ds0 = segment_end.ds1
                            segment_begin.ds1 = dsh1  # segment_begin.BG_REGION_DS  # dsh1
                            segment_begin.var_ds_ratio = ratio1
                            gap_constant_segment = MeshSegment(sh1, sh0)
                            gap_constant_segment.set_ds(dsh1, dsh0)
                            self.segment_list = numpy.insert(self.segment_list, 1, gap_constant_segment)
                            gap_segment_2 = MeshSegment(sh0, s1_begin)
                            gap_segment_2.set_ds(dsh0, ds1_begin)  # segment_begin.BG_REGION_DS
                            self.segment_list = numpy.insert(self.segment_list, 2, gap_segment_2)
                            self.segment_list[2].var_ds_ratio = ratio0
                        else:
                            print("Error: n_test0 and n_test1")

    def remove_shift(self):
        """ determine the shift and shift all segments accordingly. """
        segment_begin = self.segment_list[0]
        phi_shift = segment_begin.s0
        self.phi_shift = phi_shift
        for segments in self.segment_list:  # works for both cases: right and left shift
            segments.s0 = segments.s0 - phi_shift
            segments.s1 = segments.s1 - phi_shift
        return phi_shift

    def revert_shift(self):
        """ determine the previous shift and revert to original mesh without any shifts. """
        for segments in self.segment_list:  # works for both cases: right and left shift
            segments.s0 = segments.s0 + self.phi_shift
            segments.s1 = segments.s1 + self.phi_shift
        self.phi_shift = 0



""" sub window for tkinter_mainapp.py to adjust the theta and phi mesh segment coordinates manually. """
import tkinter as tk
import numpy as np
import matplotlib as mpl
from mesh_generator.ui.scroll_bar import VerticalScrolledFrame


class TableChart:
    """
    Table chart of mesh_box 3d coordinates. This table can be changed manually from tkinter_mainapp.py.
    The change of coordinates will also show in the matplotlib rectangle patches.
    """

    def __init__(self, root, ui_session, ax, patch_list):
        self.root = root
        self.ui_session = ui_session
        self.ax = ax
        self.patch_list = patch_list

        self.root.geometry("650x250")  # size of tkinter window.
        self.root.minsize(width=620, height=250)  # minimum size of tkinter window.
        self.root.config(background="light gray")
        self.root.wm_title("Box Chart")
        self.text_box_list = []
        self.mainFrame = VerticalScrolledFrame(self.root, width=550, height=200, borderwidth=2, relief=tk.SUNKEN,
                                               background="gray")
        self.mainFrame.pack(fill=tk.BOTH)
        tk.Button(self.root, text='Apply Changes', highlightbackground='gray', activeforeground='navy',
                  command=self.try_submit_changes).pack(side=tk.BOTTOM)

        # BOX TABLE
        self.box_table()

    def box_table(self):
        tk.Label(self.mainFrame, text='t0', relief=tk.RIDGE, bg='lightgray', bd=2, width=12).grid(row=1, column=1,
                                                                                                  sticky=tk.NSEW,
                                                                                                  columnspan=1)
        tk.Label(self.mainFrame, text='t1', relief=tk.RIDGE, bg='lightgray', bd=2, width=12).grid(row=1, column=2,
                                                                                                  sticky=tk.NSEW,
                                                                                                  columnspan=1)
        tk.Label(self.mainFrame, text='p0', relief=tk.RIDGE, bg='lightgray', bd=2, width=12).grid(row=1, column=3,
                                                                                                  sticky=tk.NSEW,
                                                                                                  columnspan=1)
        tk.Label(self.mainFrame, text='p1', relief=tk.RIDGE, bg='lightgray', bd=2, width=12).grid(row=1, column=4,
                                                                                                  sticky=tk.NSEW,
                                                                                                  columnspan=1)

        for i in range(1, len(self.ui_session.box_list) + 1):
            label = "Box # " + str(i)
            tk.Label(self.mainFrame, text=label, relief=tk.RIDGE, bg='lightgray', bd=2, width=12).grid(row=i + 1,
                                                                                                       column=0,
                                                                                                       sticky=tk.NSEW)
            rect = self.ui_session.box_list[i - 1]
            cols = []
            for j in range(4):
                e = tk.Entry(self.mainFrame, width=12)
                if j == 0 and rect.t0 is not None:
                    e.insert(tk.END, str(rect.t0))
                elif j == 1 and rect.t1 is not None:
                    e.insert(tk.END, str(rect.t1))
                elif j == 2 and rect.p0 is not None:
                    e.insert(tk.END, str(rect.p0))
                elif j == 3 and rect.p1 is not None:
                    e.insert(tk.END, str(rect.p1))
                else:
                    e.insert(tk.END, '')
                e.grid(row=i + 1, column=j + 1, sticky=tk.NSEW)
                cols.append(e)
            self.text_box_list.append(cols)

    def get_entry(self, row, row_num):
        try:
            t0 = float(row[0].get())
        except ValueError:
            t0 = None
            print("Error: t0 is not a number.")

        try:
            t1 = float(row[1].get())
        except ValueError:
            t1 = None
            print("Error: t1 is not a number.")

        try:
            p0 = float(row[2].get())
        except ValueError:
            p0 = None
            print("Error: p0 is not a number.")

        try:
            p1 = float(row[3].get())
        except ValueError:
            p1 = None
            print("Error: p1 is not a number.")

        self.check_valid(t0, t1, row_num, 't')
        self.check_valid(p0, p1, row_num, 'p')

    def check_valid(self, s0, s1, row_num, mesh_type: str):
        if s0 is not None and s1 is not None:
            if s0 < s1:
                if mesh_type == 't' and 0 <= s0 <= np.pi and 0 <= s1 <= np.pi:
                    self.ui_session.box_list[row_num - 1].t0 = s0
                    self.ui_session.box_list[row_num - 1].t1 = s1
                if mesh_type == 'p' and -2 * np.pi <= s0 <= 4 * np.pi and -2 * np.pi <= s0 <= 4 * np.pi:
                    self.ui_session.box_list[row_num - 1].p0 = s0
                    self.ui_session.box_list[row_num - 1].p1 = s1
            else:
                print("s0 needs to be less than s1.")

        if s0 is None and s1 is not None:
            if mesh_type == 't':
                if self.ui_session.box_list[row_num - 1].t0 < s1 and 0 <= s1 <= np.pi:
                    self.ui_session.box_list[row_num - 1].t1 = s1
                else:
                    print("t0 needs to be less than t1.")
            if mesh_type == 'p':
                if self.ui_session.box_list[row_num - 1].p0 < s1 and -2 * np.pi <= s1 <= 4 * np.pi:
                    self.ui_session.box_list[row_num - 1].p1 = s1
                else:
                    print("p0 needs to be less than p1.")

        if s0 is not None and s1 is None:
            if mesh_type == 't':
                if self.ui_session.box_list[row_num - 1].t1 > s0 and 0 <= s0 <= np.pi:
                    self.ui_session.box_list[row_num - 1].t0 = s0
                else:
                    print("t0 needs to be less than t1.")
            if mesh_type == 'p':
                if self.ui_session.box_list[row_num - 1].p1 > s0 and -2 * np.pi <= s0 <= 4 * np.pi:
                    self.ui_session.box_list[row_num - 1].p0 = s0
                else:
                    print("p0 needs to be less than p1.")

    def try_submit_changes(self):
        try:
            self.submit_changes()
        except:
            print("Failed to submit changes. ")

    def submit_changes(self):
        row_num = 1
        for row in self.text_box_list:
            self.get_entry(row, row_num)
            row_num += 1

        # remove rectangles from plot
        for box in self.patch_list:
            box.remove()
        self.patch_list.clear()

        # delete all text on plot
        self.ax.texts.clear()

        # # draw new boxes on the image
        ii = 1
        for rect in self.ui_session.box_list:
            rect_patch = mpl.patches.Rectangle((min(rect.p0, rect.p1), min(rect.t0, rect.t1)),
                                               np.abs(rect.p0 - rect.p1),
                                               np.abs(rect.t0 - rect.t1), fill=False)
            label_box = "Box # %s" % ii
            self.ax.text(rect.p0, rect.t0, label_box)
            self.ax.add_patch(rect_patch)
            self.patch_list.append(rect_patch)
            ii += 1

        # close the window
        self.root.destroy()
